# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for creating, chopping, concatenating and otherwise processing
strings.

The functions in this module that operate on multi-line strings have
versions that start with ``split_``, which accept iterables of lines
instead of entire strings.
"""

__all__ = [
    'hasspace', 'func_name',
    'format_list', 'parse_pair', 'parse_list',
    'align', 'split_align',
    'horiz_cat', 'split_horiz_cat',
    'make_box', 'register_box_style',
    'check_value',
    'to_casefold', 'to_lower', 'to_upper', 'to_hex',
    'camel2snake', 'snake2camel',
    'multiline_repr',
]


from collections import deque, namedtuple
from itertools import chain, islice, repeat
from math import ceil
from operator import index
from os import linesep

from .mapping import option_lookup
from .recipes import grouper


def hasspace(x):
    """
    Check if string `x` contains a space.

    If `x` is not a string, and is not Falsy, it is coerced into a
    string. Falsy inputs never have spaces (according to this function).
    """
    if not x:
        return False
    x = str(x)
    s = x.split()
    return len(s) != 1 or x != s[0]


def func_name(obj):
    """
    Retrieves the qualified function name of an object.

    The following rules are implemented:
    1. If the object is None, return None
    2. If the object has a `__qualname__` attribute, return it
    3. If the object has a `__name__` attribute, return it
    4. Return the :py:func:`repr` of the object
    """
    if obj is None:
        return None
    name = getattr(obj, '__qualname__', None)
    if name is not None:
        return name
    name = getattr(obj, '__name__', None)
    if name is None:
        return repr(obj)
    return name


def format_list(iterable, width=8, format=None, sep=', ', indent=None):
    """
    Display the elements of the iterable, `width` elements per line.

    Parameters
    ----------
    iterable : iterable
        The iterable to pretty-print.
    width : int
        The number of elements to print per line. Defaults to 8. If
        :py:obj:`None`, the entire list will appear on one line.
    format : str
        A new-style format string to represent each value. Defaults to
        ``'{}'``.
    sep : str
        The separator to place between entries on the same line.
        Defaults to ``', '``.
    indent : int or str
        If :py:class:`int`, the number of spaces by which to indent each
        line of the input. If a :py:class:`str`, act as a literal prefix
        to each line. Defaults to ``4`` unless ``width`` is
        :py:obj:`None`, in which case the default is an empty string.

    Return
    ------
    formatted_data : str
        A string containing the formatted data.
    """
    if format is None:
        format = '{}'

    if indent is None:
        indent = '' if width is None else 4

    try:
        spacer = ' ' * index(indent)
    except TypeError:
        spacer = str(indent)

    if width is None:
        result = sep.join(format.format(x) for x in iterable)
    else:
        joiner = sep.rstrip() + '\n' + spacer
        result = joiner.join(sep.join(format.format(x) for x in group)
                             for group in grouper(iterable, width))

    return spacer + result


class parse_pair:
    """
    Configure a parser for delimited pairs of items.

    Pairs get special treatment because they are a ubiquitous type
    that can be used to implement, among other things, ranges.
    """
    def __init__(self, sep='-', trim=None, item_type=None, output_type=None):
        """
        Configure a parser that can be used multiple times.

        All input parameters become attributes of the parser. `sep` is
        always stored as a sequence, as is `item_type`.

        Parameters
        ----------
        sep : str or sequence[str]
            A separator string, such as ``'-'`` or ``':'``, or a
            sequence of such strings. Any item in the sequence is a
            valid separator, and they may be mixed and matched. The
            default is ``'-'``.
        trim : callable or None
            A function to trim items after splitting them around `sep`.
            A typical value is ``str.strip``. The default is None (no
            trimming is done).
        item_type : callable, tuple[callable or None, callable or None]
                    or None
            Conversion function to apply to both or either of the items
            of the pair. May be None or a 2-tuple with either element
            None, which is equivalent to `str`. Callables should raise
            a :py:exc:`TypeError` or :py:exc:`ValueError` when invalid
            inputs are provided.
        output_type : callable or None
            How to return the pair. Callable will be invoked as
            ``output_type(first, second)``. Default is to return a
            tuple.
        """
        self.sep = (sep,) if isinstance(sep, str) else tuple(sep)
        self.trim = trim
        try:
            item_type = tuple(item_type)
        except TypeError:
            item_type = (item_type, item_type)
        else:
            if len(item_type) != 2:
                raise ValueError('item_type must be a 2-element sequence')
        self.item_type = item_type
        self.output_type = output_type

    @property
    def __name__(self):
        """
        The name of this callable.
        """
        sepstr = repr(self.sep[0]) if len(self.sep) == 1 else repr(self.sep)
        if self.item_type[0] == self.item_type[1]:
            itemstr = func_name(self.item_type[0])
        else:
            itemstr = '({}, {})'.format(func_name(self.item_type[0]),
                                        func_name(self.item_type[1]))
        return '{}(sep={}, trim={}, item_type={}, output_type={})'.format(
            type(self).__name__, sepstr, func_name(self.trim), itemstr,
            func_name(self.output_type)
        )

    def __call__(self, string):
        """
        Parse a pair according to the configuration.

        Parameters
        ----------
        string : str
            The text to parse.

        Return
        ------
        pair : tuple or self.output_type
           2-tuple or instance of whatever `self.output_type` returns.

        Raises
        ------
        ValueError
            If splitting on `self.sep` results in a number of items
            that is not a pair.
        """
        strings = [string]
        for sep in self.sep:
            strings = [s for item in strings for s in item.split(sep)]
        if len(strings) != 2:
            raise ValueError('Pair requires 2 items, '
                             'found {}'.format(len(strings)))
        if self.trim:
            strings = [self.trim(s) for s in strings]
        items = tuple(
            s if item_type is None else item_type(s)
            for item_type, s in zip(self.item_type, strings)
        )
        if self.output_type is None:
            return items
        return self.output_type(*items)

    @classmethod
    def inclusive_range(cls, sep='-', trim=None, output_type=range):
        """
        Create a parser for ranges of :py:type:`int` objects into a
        :py:type:`range` (or compatible type).

        To parse a slice with no step, use ``sep=':'`` and
        ``output_type=slice``.

        Parameters are the same as for the class, except that
        `item_type` is fixed.
        """
        return cls(sep=sep, trim=trim, item_type=(int, lambda i: int(i) + 1),
                   output_type=output_type)


class parse_list:
    """
    Configure a parser for delimited lists of items.

    Lists can contain items of different types, which are determined
    by sequentially applying a list of sub-parsers.
    """
    def __init__(self, sep=',', trim=None, item_types=None):
        """
        Configure a parser that can be used multiple times.

        All input parameters become attributes of the parser. `sep` is
        always stored as a sequence.

        Parameters
        ----------
        sep : str or sequence[str]
            A separator string, such as ``'-'`` or ``':'``, or a
            sequence of such strings. Any item in the sequence is a
            valid separator, and they may be mixed and matched. The
            default is ``'-'``.
        trim : callable or None
            A function to trim items after splitting them around `sep`.
            A typical value is ``str.strip``. The default is None (no
            trimming is done).
        item_types : callable, sequence[callable] or None
            Conversion function(s) to apply to the items in the list.
            The first function that does not raise an error is chosen,
            so the order of the list is important. Currently, only
            :py:exc:`TypeError` or :py:exc:`ValueError` are regarded as
            markers for invalid values. Add a None parser at the end of
            the list to indicate a catch-all equivalent to `str`.
        """
        self.sep = (sep,) if isinstance(sep, str) else tuple(sep)
        self.trim = trim
        try:
            item_types = tuple(item_types)
        except TypeError:
            item_types = (item_types,)
        self.item_types = item_types

    @property
    def __name__(self):
        """
        The name of this callable.
        """
        sepstr = repr(self.sep[0]) if len(self.sep) == 1 else repr(self.sep)
        if len(self.item_type) == 1:
            itemstr = func_name(self.item_types[0])
        else:
            itemstr = (
                '(' + ', '.join(func_name(t) for t in self.item_types) + ')'
            )
        return '{}(sep={}, trim={}, item_type={})'.format(
            type(self).__name__, sepstr, func_name(self.trim), itemstr
        )

    def __call__(self, string):
        """
        Parse a pair according to the configuration.

        Parameters
        ----------
        string : str
            The text to parse.

        Return
        ------
        elements : list
           List of parsed elements.

        Raises
        ------
        ValueError
        TypeError
            If an element could not be parsed. The error will be
            forwarded from the final parser.
        """
        strings = [string]
        for sep in self.sep:
            strings = [s for item in strings for s in item.split(sep)]
        if self.trim:
            strings = [self.trim(s) for s in strings]

        def parse(s):
            for item_type in self.item_types:
                if item_type is None:
                    return s
                try:
                    return item_type(s)
                except (TypeError, ValueError):
                    continue
            raise TypeError(
                'Unable to parse {!r} as any of {{{}}}'.format(
                    s, ', '.join(func_name(p) for p in self.item_types)
                )
            )

        return [parse(s) for s in strings]

    @classmethod
    def number_list(cls, sep=',', range_sep='-', trim=None):
        """
        Create a parser for a list of :py:type:`int` objects that also
        supports ranges.

        Parameters are the same as for the class, except that
        `item_type` is fixed and `range_sep` is the `sep` parameter of
        :py:class:`parse_pair`. `trim` is passed as well.
        """
        range_parser = parse_pair(sep=range_sep, trim=trim, item_type=int)
        return cls(sep=sep, trim=trim, item_types=[range_parser, int])


def _maxlen(lines, len_key=len):
    """
    Return the maximum length in an iterable based on the specified
    length function.
    """
    return max(len_key(line) for line in lines) if lines else 0


#: Mapping of acceptable alignments inputs to :py:func:`align`,
#: :py:func:`horiz_cat` and :py:func:`print_box`.
_alignments = {
    "left": '<',   "<": '<',
    "center": '^', "^": '^',
    "right": '>',  ">": '>',
    "none": None,  "": None, None: None,
}


#: Mapping of acceptable overflow handling options to :py:func:`align`.
#: The :py:obj:`None` key represents callables. The values are a
#: function that accepts an iterable of lines and a width and returns
#: the updated lines and width with overflow handled. Any overflow lines
#: are expected to be handled with the ``'skip'`` algorithm after that.
_align_overflow = {
    "extend": lambda lines, width, len, trunc: \
        (list(lines), max(width, _maxlen(lines, len))),
    "trunc": lambda lines, width, len, trunc: \
        ([line[:width] for line in lines], width),
    "skip": lambda lines, width, len, trunc: (lines, width),
    None: lambda lines, width, len, trunc: \
        ([trunc(line) for line in lines], width),
}


#: Mapping of normalized alignments (obtained from
#: :py:data:`_alignments`) to a function of ``string`` and ``delta``
#: that formats the string with additional padding adding a total of
#: ``delta`` spaces to it. ``delta`` is always expected to be positive.
_align_fn = {
    "<": lambda string, delta: '{}{}'.format(string, ' ' * delta),
    ">": lambda string, delta: '{}{}'.format(' ' * delta, string),
    "^": lambda string, delta: '{}{}{}'.format(' ' * (delta // 2), string,
                                               ' ' * ((delta + 1) // 2)),
}


def align(string, *args, **kwargs):
    """
    Align a string in a space-padded field of the specified width.

    This function differs from just doing something like ::

        {0:{1}{2}}.format(string, alignment, width)

    because the "actual" length of the string is computed by
    `len_key` rather than the builtin :py:func:`len`. It also
    correctly handles multiline strings.

    Parameters
    ----------
    string : str
        The string to align. May contain multiple lines.
    alignment : str or None
        A value that indicates how to align lines. Recognized options
        are as follows:

        ``'left'`` or ``'<'``
            `string` is padded on the right to `width`.
        ``'center'`` or ``'^'``
            `string` is padded equally on each side to `width`.
        ``'right'`` or ``'>'``
            `string` is padded on the left to `width`.
        ``'none'``, ``''`` or ``None``
            `string` is returned as-is, regardless of `width`.

    width : int or None
        The minimum field width to align in. If :py:obj:`None`, use the
        maximum line length as the field width. :py:obj:`None` does not
        make much sense for a single-line string: it will just be
        returned as-is. If `width` is shorter than *any* of the lines,
        it will have no effect.
    overflow : str or callable
        How to handle overflow lines (wider than width). The following
        options are accepted:

        ``'extend'``
            The effective width will be the largest of the maximum line
            length and `width`. This is the default.
        ``skip``
            Skip longer lines, but align shorter ones to `width`.
        ``'trunc'``
            Truncate to `width`. This may not work correctly if
            `len_key` is not the builtin function :py:func:`len`.
            Use the callable option in that case.
        ``callable``
            A function that accepts long input lines and returns a
            truncated line. This can do special truncation or any other
            operation the user desires. The result will be aligned using
            ``overflow=skip`` in this case. The callable will only be
            applied to overflow lines.

        This parameter is completely ignored if `width` is
        :py:obj:`None`.
    len_key : callable
        A custom callable for computing the lengths of strings. This can
        be useful for example if the strings contain tabs or some
        similar sequence whose display width is not necessarily the raw
        string width. Defaults to the builtin :py:func:`len`.

    Return
    ------
    aligned : str
        `string` aligned in a field of size `width` according to the
        spec. If the input string is greater than or equal to `width`
        in length, it is returned as-is.
    """
    return '\n'.join(split_align(str(string).splitlines(), *args, **kwargs))


def split_align(strings, alignment, width=None, overflow='extend',
                len_key=len):
    """
    Identical to :py:func:`align` except that the lines are passed in
    as an iterable instead of a single string.

    The result is a list of lines rather than a single string.
    """
    strings = [str(s) for s in strings]
    if width is None:
        width = _maxlen(strings, len_key)
    overflow_fn = option_lookup(
        'overflow', _align_overflow, overflow,
        key_func=lambda x: None if callable(x) else str(x).casefold()
    )
    strings, width = overflow_fn(strings, width, len_key, overflow)
    deltas = [width - len_key(string) for string in strings]

    # Normalize alignment
    alignment = option_lookup('alignment', _alignments, alignment,
                              key_func=lambda opt: str(opt).casefold())

    if not alignment:
        return strings
    formatter = _align_fn[alignment]

    strings = [formatter(string, delta) if delta >= 0 else string
               for string, delta in zip(strings, deltas)]
    return strings


#: Mapping of accepted values for :py:func:`horiz_cat`'s `missing`
#: parameter to list indices. Numerical keys get mapped to
#: :py:obj:`None`.
_hc_missing = {
    'down': None,
    'missing': None,
    'trunc': None,
    'empty': '',
    'first': 0,
    'last': -1,
    None: None,
}


def horiz_cat(*strings, **kwargs):
    """
    Concatenate multi-line strings side-by-side.

    For single line strings, this function is equivalent to
    ``prefix + sep.join(*strings) + suffix``.

    Parameters
    ---------
    *strings : str
        Any number of strings. This function is pointless if none of
        the strings have more than one line, but multiple lines are not
        a requirement.
    prefix : str
        A prefix that will be prepended to each line of the result. This
        can be used to do indentation, among other things. Default is
        empty.
    sep : str
        The separator to insert between columns. The separator will be
        omitted around empty columns if ``missing='missing'``. Default
        is a single space.
    suffix : str
        A suffix that will be added to each line in the result. Default
        is empty.
    linesep : str
        The line separator to use. The default is :py:obj:`os.linesep`.
    alignment : str, sequence[str] or None
        A value that indicates how to align strings. Acceptable
        values are as follows:

        ``'left'`` or ``'<'``
            Each string is padded on the right to the length of the
            maximum line length. Shorter lines are left aligned.
        ``'center'`` or ``'^'``
            Each string is padded equally on each side to the length of
            the maximum line length. Shorter lines are center aligned.
        ``'right'`` or ``'>'``
            Each string is padded on the left to the length of the
            maximum line length. Shorter lines are right aligned.
        ``'none'``, ``''`` or ``None``
            Strings are not padded at all, just concatenated as-is.

        If a sequence if passed in, it must contain as many elements as
        `strings`. Each column in `strings` will be aligned according to
        the corresponding alignment. Having ``'none'`` elements is not
        strictly forbidden in this case, but it may completely throw off
        the formatting of the following columns.

        The default is ``'<'``.
    missing : str or int
        A specification for how to handle strings with fewer lines than
        the others. This does not apply to empty lines within the
        string. Acceptable values are as follows:

        ``'down'``
            Shift all shorter inputs down using empty lines.
        ``'trunc'``
            Truncate all inputs to the smallest number of lines.
        ``'empty'``
            The default is to treat missing lines as through they are
            empty lines. All alignement rules apply.
        ``'last'``
            Repeat the last line.
        ``'first'``
            Repeat the first line. A synonym for ``missing=0``.
        `int` or string that evaluates to `int`
            Repeat the n-th line. Use this carefully as it will raise an
            error if any of the strings have fewer than ``n+1`` lines.

        The default is ``'empty'``.
    len_key : callable
        A custom callable for computing the lengths of strings. This can
        be useful for example if the strings contain tabs or some
        similar sequence whose display width is not necessarily the raw
        string width. Defaults to the builtin :py:func:`len`.
    titles : sequence[str] or None
        If provided, a list of column titles for each column. `titles`
        can appear with or without `headers`, which provides the
        underline. Titles behave as a normal line in `strings`.
        Multiline titles are supported.
    headers : str or callable or sequence[str or callable] or None
        Text that separates `titles` from `strings`. There are two
        options available:

        string
            Aligned the same way as each column in `strings`
        callable
            Called with the column width as a single argument, the
            return value is the column header. A common option include
            ``c.__mul__`` for a single repeated character ``c``.

        Columns are separated by `hsep` instead of `sep`. Multiline
        headers are not supported, though not technically an error.
    hsep : str or None
        If `headers` is used, column separator for the header row.
        `titles` uses `sep`. The length of `hsep` must be the same as
        that of `sep` if `headers` is used. If not set, defaults to
        `sep` when `headers` is used. Ignored if `headers` is not set.
    width : int or sequence[int] or None
        The minimum width for each column, or across all columns. If
        not set, the actual column width will be used.

    Return
    ------
    cat : str
        A string that is a side-by-side concatenation of the inputs
        given the selected options.

    Raises
    ------
    ValueError
        If `alignment` or `missing` are set to invalid values. Also if
        `sep` and `hsep` are both used and do not match in size with
        non-`None` alignment.
    IndexError
        If `missing` is an integer or a string that parses as an
        integer but one of the input strings does not contain the
        specified line index. Or if any of `alignment`, `titles`,
        `headers`, `minwidth` are sequences that are of different length
        than `strings`.
    """
    # Arrange columns as 2D list
    columns = [string.splitlines() or [''] for string in strings]
    return split_horiz_cat(*columns, **kwargs)


def split_horiz_cat(*columns, prefix='', sep=' ', suffix='', linesep=linesep,
                    alignment='<', missing='empty', len_key=len, titles=None,
                    headers=None, hsep=None, width=None):
    """
    Identical to :py:func:`horiz_cat`, except the inputs are sequences
    of strings already split into lines.

    If `linesep` is :py:obj:`None`, the output will not be combined
    into a string, but will be returned as a :py:class:`list` instead.
    """
    maxlines = _maxlen(columns, len_key=len)

    # Handle missing elements
    def missing_key(option):
        try:
            int(missing)
        except:
            if isinstance(missing, str):
                return missing.casefold()
            return False  # Known to be invalid key
        else:
            return None

    def missing_value(option, key, value):
        down = False
        if key is None:
            fn = lambda col: col[int(option)]
        elif key == 'down':
            down = True
            fn = lambda col: ''
        elif key == 'trunc':
            return list(map(list, columns))
        elif isinstance(value, int):
            fn = lambda col: col[value]
        else:
            fn = lambda col: value

        def args(col):
            rep = repeat(fn(col), maxlines - len(col))
            if down:
                return rep, col
            return col, rep
        return [list(chain(*args(col))) for col in columns]

    n = len(columns)

    def check_len(arg, name, t=str, c=False):
        if (c and callable(arg)) or isinstance(arg, (type(None), t)):
            return [arg] * n
        if len(arg) != n:
            raise IndexError('Number of {} ({}) does not match number of '
                             'columns ({})'.format(name, len(arg), n))
        return arg

    alignment = check_len(alignment, 'alignments')
    titles = check_len(titles, 'titles')
    headers = check_len(headers, 'headers', c=True)
    widths = check_len(width, 'widths', t=int)

    titles = [[] if t is None else t.splitlines() or [''] for t in titles]
    title_height = max(map(len, titles), default=0)
    titles = [('\n' * (title_height - len(t))).splitlines() + t for t in titles]

    include_headers = any(h is not None for h in headers)
    if include_headers:
        header_strs = [[h] if isinstance(h, str) else [''] for h in headers]
    else:
        header_strs = [[]] * n

    if hsep is None:
        hsep = sep
    elif include_headers and \
            len(columns) > 1 and \
            len_key(hsep) != len_key(sep):
        raise ValueError('`hsep` and `sep` must be the same length')

    columns = option_lookup('missing', _hc_missing, missing,
                            key_func=missing_key, value_func=missing_value)
    columns = [split_align(title + header + column,
                           alignment=align, width=width, len_key=len_key)
               for title, header, align, width, column in
                   zip(titles, header_strs, alignment, widths, columns)]

    line_zip = zip(*columns)
    lines = [prefix + sep.join(items) + suffix
             for items in islice(line_zip, title_height)]
    if include_headers:
        lines.append(prefix + hsep.join(h(len(s)) if callable(h) else s
                     for h, s in zip(headers, next(line_zip))) + suffix)
    lines += [prefix + sep.join(items) + suffix for items in line_zip]

    if linesep is None:
        return lines
    return linesep.join(lines)


#: Styles available to :py:func:`make_box`, keyed by name.
_make_box_styles = {}


#: The values of :py:data:`_make_box_styles`.
_make_box_style = namedtuple('_make_box_style', [
    'top', 'left', 'bottom', 'right', 'ul', 'ur', 'bl', 'br'])


def make_box(string, style='ascii-block', alignment='^', linesep=linesep,
             horizontal_padding=1, vertical_padding=0, len_key=len):
    """
    Surrounded the input string by a box.

    Parameters
    ----------
    string : str
        A string, which may contain multiple lines.
    style : str
        The type of box to draw. Styles are registered with
        :py:func:`register_box_style`. Predefined styles are

        - ``'ascii-block'``
        - ``'ascii-line'``
        - ``'shaded'``
        - ``'block'``
        - ``'half-block'``
        - ``'half-block-inner'``
        - ``'line'``
        - ``'bold-line'``
        - ``'rounded-line'``
        - ``'double-line'``

        All but the ``'ascii-*'`` styles use unicode characters. The
        default is ``'ascii-block'``.
    alignment : str or None
        A value that indicates how to align lines in multiline strings.
        Acceptable values are as follows:

        ``'left'`` or ``'<'``
            Each string is padded on the right to the length of the
            maximum line length. Shorter lines are left aligned.
        ``'center'`` or ``'^'``
            Each string is padded equally on each side to the length of
            the maximum line length. Shorter lines are center aligned.
        ``'right'`` or ``'>'``
            Each string is padded on the left to the length of the
            maximum line length. Shorter lines are right aligned.

        Default is ``'^'``.
    linesep : str
        The line separator to use. The default is :py:obj:`os.linesep`.
    horizontal_padding : int
        The number of spaces to place between the string and the left
        and right borders. Defaults to one space on each side.
    vertical_padding : int
        The number of newlines to place between the string and the top
        and bottom border. Defaults to one line on each side.
    len_key : callable
        A custom callable for computing the lengths of strings. This can
        be useful for example if the strings contain tabs or some
        similar sequence whose display width is not necessarily the raw
        string width. Defaults to the builtin :py:func:`len`.

    Return
    ------
    box : str
        The input string surrounded by a border box. The return value
        will always be a multi-line string.
    """
    # Get the strings
    vertical_padding = [' '] * vertical_padding
    horizontal_padding  = ' ' * horizontal_padding
    lines = vertical_padding + string.splitlines() + vertical_padding
    lines = split_align(lines, alignment=alignment, width=None,
                        len_key=len_key)
    lines = [horizontal_padding + line + horizontal_padding for line in lines]

    # Get left and right styles, taking into account the corner styles
    style = option_lookup('style', _make_box_styles, style,
                          key_func=str.casefold)

    lwidth = max(_maxlen(style.ul, len_key), _maxlen(style.bl, len_key),
                 _maxlen(style.left, len_key))
    rwidth = max(_maxlen(style.ur, len_key), _maxlen(style.br, len_key),
                 _maxlen(style.right, len_key))
    cwidth = _maxlen(lines, len_key)

    cheight = len(lines)

    ul = split_align(style.ul, alignment='>', width=lwidth, len_key=len_key)
    bl = split_align(style.bl, alignment='>', width=lwidth, len_key=len_key)
    left = split_align(style.left * ceil(cheight / len(style.left)),
                       alignment='>', width=lwidth)
    right = split_align(style.right * ceil(cheight / len(style.right)),
                        alignment='none', width=rwidth)
    top = split_horiz_cat(*([style.top] * ceil(cwidth / _maxlen(style.top))),
                          alignment='<', sep='', linesep=None, missing='down',
                          len_key=len_key)
    top = [line[:cwidth] for line in top]
    top = split_horiz_cat(ul, top, style.ur, sep='', linesep=None,
                          alignment='none', missing='empty', len_key=len_key)
    bottom = split_horiz_cat(
        *([style.bottom] * ceil(cwidth / _maxlen(style.bottom))), sep='',
        alignment='<', linesep=None, missing='empty', len_key=len_key
    )
    bottom = [line[:cwidth] for line in bottom]
    bottom = split_horiz_cat(bl, bottom, style.br, sep='', linesep=None,
                             alignment='none', missing='empty',
                             len_key=len_key)

    lines = split_horiz_cat(left, lines, right, sep='', linesep=None,
                            alignment='none', len_key=len_key)
    lines = top + lines + bottom

    return linesep.join(lines)


def register_box_style(name, top, left, bottom, right, ul, ur, bl, br):
    """
    Create a new box style that is available to :py:func:`make_box`
    through `name`.

    A style is defined by specifying the characters to use for each side
    and corner of the box. Normally, borders are specified as a single
    character, but this is not required.

    Registering an existing style will silently overwrite the previous
    style.

    Parameters
    ----------
    name : str
        The name by which the style will be accessible to
        :py:func:`make_box` via the `style` parameter.
    top : str
        The character or characters to use for the top side of the box.
    left : str
        The character or characters to use for the left side of the box.
    bottom : str
        The character or characters to use for the bottom side of the
        box.
    right : str
        The character or characters to use for the right side of the
        box.
    ul : str
        The character or characters that will join `top` and `left`
        borders in the upper left-hand corner.
    ur : str
        The character or characters that will join `top` and `right`
        borders in the upper right-hand corner.
    bl : str
        The character or characters that will join `bottom` and
        `left` borders in the lower left-hand corner.
    br : str
        The character or characters that will join `bottom` and
        `right` borders in the lower right-hand corner.


    The elements of the left column of the box will be right-aligned
    based on the longest among `ul`, `left` and `bl`. The elements
    of the right side will be similarly left-aligned based on the
    longest among `ur`, `right` and `br`.

    Style string elements may contain multiple lines. In that case, all
    the lines will be left-aligned.
    """
    _make_box_styles[name.casefold()] = _make_box_style(
        top=top.splitlines(), left=left.splitlines(),
        bottom=bottom.splitlines(), right=right.splitlines(),
        ul=ul.splitlines(), ur=ur.splitlines(),
        bl=bl.splitlines(), br=br.splitlines()
    )


register_box_style('ascii-block', *('#' * 8))
register_box_style('ascii-line', *'-|-|', *('+' * 4))

register_box_style('shaded',
    top='\u2593', left='\u2593', bottom='\u2593\n\u2592', right='\u2593\u2592',
    ul='\u2593', ur='\u2593\u2591', bl='\u2593\n\u2591',
    br='\u2593\u2592\n\u2592\u2592')
register_box_style('block', *('\u2588' * 8))
register_box_style('half-block',
    top='\u2580', bottom='\u2584', left='\u258C', right='\u2590',
    ul='\u259B', ur='\u259C', bl='\u2599', br='\u259F')
register_box_style('half-block-inner',
    top='\u2584', bottom='\u2580', left='\u2590', right='\u258C',
    ul='\u2597', ur='\u2596', bl='\u259D', br='\u2598')

register_box_style('line',
    top='\u2500', bottom='\u2500', left='\u2502', right='\u2502',
    ul='\u250C', ur='\u2510', bl='\u2514', br='\u2518')
register_box_style('rounded-line',
    top='\u2500', bottom='\u2500', left='\u2502', right='\u2502',
    ul='\u256D', ur='\u256E', bl='\u2570', br='\u256F')

register_box_style('bold-line',
    top='\u2501', bottom='\u2501', left='\u2503', right='\u2503',
    ul='\u250F', ur='\u2513', bl='\u2517', br='\u251B')
register_box_style('bold-hline',
    top='\u2501', bottom='\u2501', left='\u2502', right='\u2502',
    ul='\u250D', ur='\u2511', bl='\u2515', br='\u2519')
register_box_style('bold-vline',
    top='\u2500', bottom='\u2500', left='\u2503', right='\u2503',
    ul='\u250E', ur='\u2512', bl='\u2516', br='\u251A')
register_box_style('bold-tline',
    top='\u2501', bottom='\u2500', left='\u2502', right='\u2502',
    ul='\u250D', ur='\u2511', bl='\u2514', br='\u2518')
register_box_style('bold-bline',
    top='\u2500', bottom='\u2501', left='\u2502', right='\u2502',
    ul='\u250C', ur='\u2510', bl='\u2515', br='\u2519')
register_box_style('bold-lline',
    top='\u2500', bottom='\u2500', left='\u2503', right='\u2502',
    ul='\u250E', ur='\u2510', bl='\u2516', br='\u2518')
register_box_style('bold-rline',
    top='\u2500', bottom='\u2500', left='\u2502', right='\u2503',
    ul='\u250C', ur='\u2512', bl='\u2514', br='\u251A')

register_box_style('double-line',
    top='\u2550', bottom='\u2550', left='\u2551', right='\u2551',
    ul='\u2554', ur='\u2557', bl='\u255A', br='\u255D')
register_box_style('double-hline',
    top='\u2550', bottom='\u2550', left='\u2502', right='\u2502',
    ul='\u2552', ur='\u2555', bl='\u2558', br='\u255B')
register_box_style('double-vline',
    top='\u2500', bottom='\u2500', left='\u2551', right='\u2551',
    ul='\u2553', ur='\u2556', bl='\u2559', br='\u255C')
register_box_style('double-tline',
    top='\u2550', bottom='\u2500', left='\u2502', right='\u2502',
    ul='\u2552', ur='\u2555', bl='\u2514', br='\u2518')
register_box_style('double-bline',
    top='\u2500', bottom='\u2550', left='\u2502', right='\u2502',
    ul='\u250C', ur='\u2510', bl='\u2558', br='\u255B')
register_box_style('double-lline',
    top='\u2500', bottom='\u2500', left='\u2551', right='\u2502',
    ul='\u2553', ur='\u2510', bl='\u2559', br='\u2518')
register_box_style('double-rline',
    top='\u2500', bottom='\u2500', left='\u2502', right='\u2551',
    ul='\u250C', ur='\u2556', bl='\u2514', br='\u255C')

register_box_style('dashed-line',
    top='\u254C', bottom='\u254C', left='\u254E', right='\u254E',
    ul='\u250C', ur='\u2510', bl='\u2514', br='\u2518')
register_box_style('dash-dotted-line',
    top='\u2504', bottom='\u2504', left='\u2506', right='\u2506',
    ul='\u250C', ur='\u2510', bl='\u2514', br='\u2518')
register_box_style('dotted-line',
    top='\u2508', bottom='\u2508', left='\u250A', right='\u250A',
    ul='\u250C', ur='\u2510', bl='\u2514', br='\u2518')

register_box_style('dashed-bold-line',
    top='\u254D', bottom='\u254D', left='\u254F', right='\u254F',
    ul='\u250F', ur='\u2513', bl='\u2517', br='\u251B')
register_box_style('dash-dotted-bold-line',
    top='\u2505', bottom='\u2505', left='\u2507', right='\u2507',
    ul='\u250F', ur='\u2513', bl='\u2517', br='\u251B')
register_box_style('dotted-bold-line',
    top='\u2509', bottom='\u2509', left='\u250B', right='\u250B',
    ul='\u250F', ur='\u2513', bl='\u2517', br='\u251B')


def check_value(string, options, insensitive=True, label='option'):
    """
    Verifiy that string is within the allowed options.

    Raise a :py:exc:`ValueError` if not found.

    Parameters
    ----------
    string : str
        The string to verify.
    options : mapping[str, str] or iterable[str]
        A container that supports the ``in`` operator. If the container
        is a mapping that can be indexed by the input, the result is
        the "normalized" version of the string that will be returned.
    insensitive : bool
        Whether or not to do a case insensitive comparison (using
        :py:meth:`str.casefold`). If :py:obj:`True`, `options` is
        expected to contain casefolded items.
    label : str
        The label used to describe `string` if it is not found in
        `options`.

    Return
    ------
    check : str
        The input string if it is in `options`. If options is a
        mapping, the result will be ``options[string]`` instead. This
        allows normalization of the input so that multiple input options
        can represent the same output value.
    """
    check = string.casefold() if insensitive else string
    if check not in options:
        raise ValueError('No such {} "{}"'.format(label, string))
    try:
        return options[check]
    except TypeError:
        return check


def to_casefold(string, conv=str):
    """
    Convert the input to a string and casefold it.

    Conversion is done by `conv`, which is normally :py:class:`str`, but
    :py:func:`repr` is sometimes a good choice as well.
    """
    return conv(string).casefold()


def to_lower(string, conv=str):
    """
    Convert the input to a string and lowercase it.

    Conversion is done by `conv`, which is normally :py:class:`str`, but
    :py:func:`repr` is sometimes a good choice as well.
    """
    return conv(string).lower()


def to_upper(string, conv=str):
    """
    Convert the input to a string and uppercase it.

    Conversion is done by `conv`, which is normally :py:class:`str`, but
    :py:func:`repr` is sometimes a good choice as well.
    """
    return conv(string).upper()


def to_hex(b, prefix='', sep=''):
    """
    Convert a string of bytes to a hex string.

    Parameters
    ----------
    b : bytes or bytearray
        The bytes to display.
    prefix : str
        The prefix to prepend to each byte. The default is an empty
        string. Another common choice is ``'0x'``.
    sep : str, optional
        The separator to place between bytes. The default is an empty
        string: hex values are concatenated all together.

    Returns
    -------
    str
        A string consisting of the characters 0-9, A-Z (as well as
        `prefix` and `sep`), with two digits per byte of input.
    """
    return sep.join(f'{x:02X}' for x in b)


def camel2snake(string):
    """
    Convert a string to snake_case, assuming input in CamelCase.

    Parameters
    ----------
    string : str
        The string to convert.

    Returns
    -------
    str
        A string similar to the input, but any uppercase letters are
        lowercased and an underscore is prepended, unless there is one
        there already.
    """
    chars = deque()
    for i, c in enumerate(str(string)):
        if c.isupper():
            if i > 0 and chars[-1] != '_':
                chars.append('_')
            c = c.lower()
        chars.append(c)
    return ''.join(chars)


def snake2camel(string, first_upper=False):
    """
    Convert a string to CamelCase, assuming input in snake_case.

    Parameters
    ----------
    string : str
        The string to convert.
    first_upper : bool
        Whether or not to capitalize the first letter.

    Returns
    -------
    str
        A string similar to the input, but any underscores removed, and
        the following letters uppercased.
    """
    chars = deque()
    next_upper = False
    for c in str(string):
        if c == '_':
            next_upper = True
            continue
        if next_upper:
            c = c.upper()
        chars.append(c)
        next_upper = False
    if first_upper:
        chars[0] = chars[0].upper()
    return ''.join(chars)


def multiline_repr(args=(), kws=()):
    """
    Decorator providing a "fancy" version of `__repr__` to classes.

    The version of `__repr__` prints the type name, with an argument
    list determined by `repr_args` and `repr_kws`. The argument lists
    are stored in class properties named `_multiline_repr_args` and
    `_multiline_repr_kws`. These aggregate across inheritance trees.

    Setting either iterable explicitly to `None` instead of an empty
    iterable will dynamically use `list(vars(self))`.

    Parameters
    ----------
    args : Iterable
        An iterable of strings or 2-element tuples (may be mixed).
        Tuples are of the form `(keyword, function)`. `function` must
        have a signature like `getattr`. If `keyword` starts with an
        asterisk (`*`), the function is expected to return an iterable
        of individual arguments, otherwise it returns a single argument.
    kws : Iterable
        An iterable of the same type as `args`, except that
        star-arguments are forbidden: each keyword may only refer to a
        single object.

    Return
    ------
    A decorator that will add a `__repr__` method and two properties
    to a class. The result of `__repr__` will be the class name and a
    pretty-printed argument list, displayed across multiple lines.

    Examples
    --------
    Indentation works with nested multiline representations::

        >>> @multiline_repr('p', 'q')
        ... class Inner:
        ...     def __init__(self, p, q=None):
        ...         self.p = p
        ...         self.q = q
        >>> @multiline_repr(['a', 'b'], ['x', ('y', lambda obj, attr: obj.z)])
        ... class Test:
        ...     def __init__(self):
        ...         self.a = 1
        ...         self.b = 2
        ...         self.x = Inner('a', 'b')
        ...         self.z = 'omega'
        ...
        >>> print(Test())
        Test(1,
             2,
             x=Inner('a',
                     q='b'),
             y='omega')

    And with inheritance::

        >>> @multiline_repr(['r', ('*s', lambda obj, attr: ('cat', 'dog'))], ['t'])
        ... class Outer(Inner):
        ...     def __init__(self):
        ...         super().__init__('start')
        ...         self.r = 'stop'
        ...         self.t = 'none'
        >>> print(Outer())
        Outer('start',
              'stop',
              'cat',
              'dog',
              q=None,
              t='none')
    """
    def get_attr(obj, attr):
        if attr.startswith('*'):
            yield from getattr(obj, attr[1:])
        else:
            yield getattr(obj, attr)

    def tidy_up(arg_list, allow_star):
        if arg_list is None:
            return None
        args = []
        for descr in arg_list:
            if isinstance(descr, str):
                func = None
            elif not isinstance(descr, tuple) or len(descr) != 2 \
                                    or not isinstance(descr[0], str):
                raise TypeError('Illegal argument descriptor: '
                                'must be str or 2-element tuple')
            else:
                descr, func = descr
            if allow_star:
                if func is None:
                    func = get_attr
            else:
                if descr.startswith('*'):
                    raise ValueError('Only positoinal arguments may have a *')
                if func is None:
                    func = getattr
            args.append((descr, func))
        return tuple(args)

    multiline_repr_args = tidy_up(args, True)
    multiline_repr_kws = tidy_up(kws, False)

    def s(k):
        return ' ' * (len(k) + 1)

    def __repr__(self):
        t = type(self).__name__
        sep1 = '\n' + s(t)
        arg_strs = (sep1.join(repr(arg).splitlines())
                        for name, func in self._multiline_repr_args
                            for arg in func(self, name))
        kw_strs = (f'{name}=' + (sep1 + s(name)).join(
                                repr(func(self, name)).splitlines()
                       ) for name, func in self._multiline_repr_kws)
        sep2 = ',' + sep1
        return (f'{t}({sep2.join(chain(arg_strs, kw_strs))})')

    def decorator(cls):
        @property
        def _multiline_repr_args(self):
            args = dict.fromkeys(vars(self), getattr).items() \
                    if multiline_repr_args is None else multiline_repr_args
            parent = super(cls, self)
            prev = parent._multiline_repr_args \
                    if hasattr(parent, '_multiline_repr_args') else ()
            return prev + args

        @property
        def _multiline_repr_kws(self):
            kws = dict.fromkeys(vars(self), getattr).items() \
                    if multiline_repr_kws is None else multiline_repr_kws
            parent = super(cls, self)
            prev = parent._multiline_repr_kws \
                    if hasattr(parent, '_multiline_repr_kws') else ()
            return tuple(prev) + tuple(kws)

        cls._multiline_repr_args = _multiline_repr_args
        cls._multiline_repr_kws = _multiline_repr_kws
        cls.__repr__ = __repr__
        return cls

    return decorator



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
:py:mod:`haggis` is the root package of the haggis library.

The root package contains the :py:data:`__version__`, a
:py:data:`Sentinel` singleton and its associated type.

The sub-packages and sub-modules in this library are arranged mostly by
category. Some of the dependencies to the various types of utilities are
optional, and the corresponding modules will only work fully if the
dependencies are present. See the :ref:`installation-extras` section in
the :ref:`installation`.
"""

__all__ = ['__version__', 'Sentinel', 'SentinelType']


from .version import __version__


class SentinelType:
    """
    A class that can be used to create sentinel objects for cases where
    :py:obj:`None` is not suitable for some reason.

    This class's truth value is always :py:obj:`False`. It does not
    allow any additional attributes to be added.

    Simply creating an empty :py:class:`object` is fine in most cases.
    """
    __slots__ = ()

    def __bool__(self):
        """
        Ensure that instances are always falsy.
        """
        return False


#: A sentinel object that can be used when :py:obj:`None` is not a
#: suitable option (e.g., when :py:obj:`None` has a special meaning).
#:
#: This object evaluates to boolean :py:obj:`False`.
Sentinel = SentinelType()


def _display_missing_extra(extra, libs=None):
    """
    Print a message to standard error regarding a missing extra.
    """
    from sys import stderr

    print(
        'This feature is only enabled with the [{extra}] extra. Try\n\n'
        'pip install haggis[{extra}]'.format(extra=extra), file=stderr
    )
    if libs:
        if isinstance(libs, str):
            libs = [libs]
        print('\nor install {} manually'.format(
            ', '.join(str(lib) for lib in libs)), file=stderr)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Various written-language related routines pertaining to numbers.
"""

__all__ = ['as_base', 'digit_count', 'english', 'metric_prefix']


from string import ascii_uppercase, digits as ascii_digits
from math import log, log10, ceil, floor
from operator import index


def as_base(n, base=10, letters=True, sign=True):
    """
    Convert integer `n` to representation as `base`.

    For bases 36 and under, digits 10 or larger can be represented by
    English letters in range A-Z. For larger bases, the output must be
    a list.

    Only absolute value of the number is coverted. For string
    representations, a ``-`` symbol can be prepended. Otherwise, it is
    the user's responsibility to handle sign.

    Parameters
    ----------
    n : int
        The number to represent.
    base : int
        The base of representation. Must be a positive integer.
        Special case of 1 is allowed.
    letters : bool, optional
        If True, represent digits larger than 9 with ASCII uppercase
        letters and return a string. If ``base > 36``, this parameter
        is ignored (implicitly False).
    sign : bool, optional
        Prepend a minus sign if returning a string and `n` is negative.
        Ignored if ``letters is False`` or ``base > 36``.

    Return
    ------
    num : str or list[int]
        If ``letters is True`` and ``base <= 36``, this is a string
        representation of `num` in `base` with optional sign.
        Otherwise, it is a list of digits from highest to lowest,
        ignoring sign.

    Notes
    -----
    Inspired by https://stackoverflow.com/a/28666223/2988730
    """
    base = int(base)
    if base < 1:
        raise ValueError('Illegal negative base')

    n = int(n) if isinstance(n, str) else index(n)
    if n < 0:
        s = True
        n = -n
    else:
        s = False

    if n == 0:
        digits = [0]
    elif base == 1:
        digits = [1] * abs(n)
    else:
        digits = []
        while n:
            digits.append(n % base)
            n //= base
        digits = digits[::-1]

    if base <= 36 and letters:
        digits = ''.join(ascii_uppercase[d - 10]
                                if d >= 10 else ascii_digits[d]
                                       for d in digits)
        if sign and s:
            digits = '-' + digits

    return digits


def digit_count(n, base=10):
    """
    Compute the number of digits required to represent an integer in a
    given base, ignoring sign.

    All numbers have at least one digit except zero. The sign is
    removed when counting digits.

    Parameters
    ----------
    n : int
        A number.
    base : int
        The base to compute the digit count of `n` in.

    Return
    ------
    The number of digits in `n` when represented in `base`.

    Notes
    -----
    For string representations, count zero as a digit using::

        max(digit_count(n, base), 1)

    """
    n = index(n)
    base = index(base)
    if base == 1:
        return n
    return ceil(log(n + 1, base))


#: Written representation of numbers [0, 10).
_basic_ones = [
    'zero', 'one', 'two', 'three', 'four',
    'five', 'six', 'seven', 'eight', 'nine'
]

#: Written representation of numbers [10, 20).
_basic_teens = [
    'ten', 'eleven', 'twelve', 'thirteen', 'fourteen',
    'fifteen', 'sixteen', 'seventeen', 'eighteen', 'nineteen'
]

#: Written representation of multiples of ten less than 100.
_basic_tens = [
    'twenty', 'thirty', 'forty', 'fifty',
    'sixty', 'seventy', 'eighty', 'ninety',
]

#: Written representation of powers of thousands up to 10**63, based on
#: the short scale most common in modern English usage (no milliard or
#: billiard): https://en.wikipedia.org/wiki/Names_of_large_numbers
_basic_powers = [
    '', 'thousand', 'million', 'billion', 'trillion', 'quadrillion',
    'quintillion', 'sextillion', 'septillion', 'octillion', 'nonillion',
    'decillion', 'undecillion', 'duodecillion', 'tredecillion',
    'quattuordecillion', 'quindecillion', 'sexdecillion', 'septendecillion',
    'octodecillion', 'novemdecillion', 'vigintillion',
]

#: Special ordinals that are not just their written number suffixed by
#: ``-th``.
_ordinal_specials = {
    1: 'first', 2: 'second', 3: 'third', 5: 'fifth', 8: 'eighth', 9: 'ninth',
    12: 'twelfth',
}

#: Special suffixes for ordinals that are not just ``-th``.
_suffix_specials = {1: 'st', 2: 'nd', 3: 'rd',}

#: The default ordinal suffix is ``-th``.
_default_suffix = 'th'


def english(num, mode='basic', hyphens=True, one_is_a=False, format='d'):
    r"""
    Convert non-negative integer into its (mostly British) English
    equivalent.

    Integers up to 999 vigintillion (10\ :sup:`64` - 1) are supported.

    Usage is modifiable British as specced out in
    http://english.stackexchange.com/a/111837/207127.

    Optional hyphens can be turned off by setting `hyphens` to
    :py:obj:`False`.

    The more American prefix "one" is preferred when the highest power
    of 10 is a unit. True British usage can be enabled be setting
    `one_is_a` to :py:obj:`True`. This will make 100 translate as "a
    hundred" instead of the (default) Americanized "one hundred".

    Supported values for `mode` are:

      - ``'basic'``: Convert ``1`` into ``'one'``.
      - ``'ordinal'``: Convert ``1`` into ``'first'``.
      - ``'suffix'``: Convert ``1`` into ``'1st'``. Usage rules are
        irrelevant with this option.

    Usage of higher powers of 10 are Americanized as well:

      - 10\ :sup:`6`: million
      - 10\ :sup:`9`: billion
      - 10\ :sup:`12`: trillion
      - 10\ :sup:`15`: quadrillion
      - etc...

    `format` is an optional string that specifies an integer format
    conforming to the Python :ref:`formatspec` (used by
    :py:class:`string.Formatter` and :py:meth:`str.format`). The default
    is ``'d'``.
    """
    # Check that it is really an integer and non-negative
    num = int(num)
    if num < 0:
        raise ValueError('Only positive integers supported')

    # If need suffix, don't check range since everything > 20 is 'th'
    if mode == 'suffix':
        teens = num % 100
        if teens > 20:
            teens %= 10
        suffix = _suffix_specials.get(teens, _default_suffix)
        return '{:{}}{}'.format(num, format, suffix)

    if mode not in ('basic', 'ordinal'):
        raise ValueError('Unsupported mode "{}"'.format(mode))

    if num < 10:
        if mode == 'basic':
            return _basic_ones[num]
        return _ordinal_specials.get(num, _basic_ones[num] + _default_suffix)

    # Check range if not in suffix mode
    max_power = 3 * len(_basic_powers)
    # Originally tried to do log10(num) >= max_power. Didn't work because
    # log10(999999999999999999) == 18.0 due to limited precision of floats
    if num >= 10**max_power:
        raise ValueError('{} is out of range. Only numbers '
                         'up to 10**{} are supported.'.format(num, max_power))

    teen_sep = '-' if hyphens else ' '

    def separate(upper, lower, sep=' '):
        """
        Add the specified separator between two strings only if they are both
        non-empty.

        If either is empty, just returns the other one.
        """
        if upper and lower:
            return upper + sep + lower
        return upper + lower

    def process_thousand(num, use_a=False, add_suffix=False):
        """
        Convert the lowest three digits of a number into English
        format.

        This method does not process zero. If `top` is `True` and the
        number is either 1 or 100-199, the prefix will be 'a' instead
        of 'one'. Do not ask this method to process 1 if the entire
        number is 1.

        Ordinal mode can be handled here as well. If the suffix is
        requested, it is generated based on the last two digits if they
        are nonzero, added as 'th' if they are.
        """
        hundreds = num // 100
        if hundreds == 1 and use_a:
            upper = 'a hundred'
        elif hundreds:
            upper = _basic_ones[hundreds] + ' hundred'
        else:
            upper = ''

        def get_one(num):
            """
            Convert a number 0-9 to text.

            Zero becomes an empty string.
            """
            if num == 0:
                return ''
            if add_suffix:
                if num in _ordinal_specials:
                    return _ordinal_specials[num]
                return _basic_ones[num] + _default_suffix
            return _basic_ones[num]

        def get_teen(num):
            """ Convert a numbers 10-19 to text. """
            if add_suffix:
                if num in _ordinal_specials:
                    return _ordinal_specials[num]
                return _basic_teens[num - 10] + _default_suffix
            return _basic_teens[num - 10]

        def get_regular(num):
            """ Convert numbers 20-99 to text. """
            ones = num % 10
            base = _basic_tens[num // 10 - 2]
            if ones:
                return base + teen_sep + get_one(ones)
            else:
                if add_suffix:
                    return base[:-1] + 'ie' + _default_suffix
                return base

        teens = num % 100
        if teens == 1 and not hundreds and use_a:
            lower = 'a'
        elif teens < 10:
            lower = get_one(teens)
        elif teens < 20:
            lower = get_teen(teens)
        else:
            lower = get_regular(teens)

        return separate(upper, lower, ' and ')

    result = ''
    # Split number into groups of three digits. ceil(log1000(num + 1)) is the
    # total number of groups of 3 digits. The +1 ensures that exact powers of
    # 1000 get punted into the next group count. Also, it is not an issue that
    # this will not work correctly for zero since that special case is handled
    # above anyway.
    digit_list = [
        (num // 1000**n) % 1000 for n in range(int(ceil(log10(num + 1) / 3)))
    ]
    # What is the last index?
    max_index = len(digit_list) - 1
    # How much has been accumulated back
    total_so_far = 0
    for index, (power, digits) in enumerate(zip(_basic_powers, digit_list)):
        if digits:
            sep = ' and ' if total_so_far < 100 else ', '
            use_a = (bool(one_is_a) and index == max_index)
            add_suffix = (mode == 'ordinal' and total_so_far == 0)
            words = separate(process_thousand(digits, use_a, add_suffix), power)
            result = separate(words, result, sep)
        total_so_far += 1000**index * digits
    return result


#: Prefixes for the powers of 10 between -24 and +24:
#: https://en.wikipedia.org/wiki/Metric_prefix
_metric_prefixes = {
    24: ('Y', 'yotta'),
    21: ('Z', 'zetta'),
    18: ('E', 'exa'),
    15: ('P', 'peta'),
    12: ('T', 'tera'),
     9: ('G', 'giga'),
     6: ('M', 'mega'),
     3: ('k', 'kilo'),
     2: ('h', 'hecto'),
     1: ('da', 'deca'),
     0: ('', ''),
    -1: ('d', 'deci'),
    -2: ('c', 'centi'),
    -3: ('m', 'milli'),
    -6: ('μ', 'micro'),
    -9: ('n', 'nano'),
   -12: ('p', 'pico'),
   -15: ('f', 'femto'),
   -18: ('a', 'atto'),
   -21: ('z', 'zepto'),
   -24: ('y', 'yocto'),
}


def metric_prefix(num, long=False, eng=False):
    r"""
    Return a number and the letter that represents its metric prefix.

    Prefixes are recognized in powers of 10\ :sup:`3` between
    10\ :sup:`-24` and 10\ :sup:`24`. Prefixes for 10\ :sup:`-2`,
    10\ :sup:`-1`, 10\ :sup:`1` and 10\ :sup:`2` are also recognized if
    ``eng=False``.

    Known prefixes are given in the following table:

    +--------+--------+--------------+
    | Prefix | Symbol | Power of 10  |
    +========+========+==============+
    | yotta  | Y      |           24 |
    +--------+--------+--------------+
    | zetta  | Z      |           21 |
    +--------+--------+--------------+
    | exa    | E      |           18 |
    +--------+--------+--------------+
    | peta   | P      |           15 |
    +--------+--------+--------------+
    | tera   | T      |           12 |
    +--------+--------+--------------+
    | giga   | G      |            9 |
    +--------+--------+--------------+
    | mega   | M      |            6 |
    +--------+--------+--------------+
    | kilo   | k      |            3 |
    +--------+--------+--------------+
    | hecto  | h      |            2 |
    +--------+--------+--------------+
    | deca   | da     |            1 |
    +--------+--------+--------------+
    | <None> | <None> |            0 |
    +--------+--------+--------------+
    | deci   | d      |           -1 |
    +--------+--------+--------------+
    | centi  | c      |           -2 |
    +--------+--------+--------------+
    | milli  | m      |           -3 |
    +--------+--------+--------------+
    | micro  | μ      |           -6 |
    +--------+--------+--------------+
    | nano   | n      |           -9 |
    +--------+--------+--------------+
    | pico   | p      |          -12 |
    +--------+--------+--------------+
    | femto  | f      |          -15 |
    +--------+--------+--------------+
    | atto   | a      |          -18 |
    +--------+--------+--------------+
    | zepto  | z      |          -21 |
    +--------+--------+--------------+
    | yocto  | y      |          -24 |
    +--------+--------+--------------+

    Parameters
    ----------
    num : number
        The number to normalize
    long : bool
        Whether to return the prefix or just the symbol. Defaults to
        :py:obj:`False` (just the symbol).
    eng : bool
        Whether to use engineering notation (omit powers that aren't)
        multiples of 3. Deaults to :py:obj:`False`, so ``centi``,
        ``deca``, etc., are viable options.

    Return
    ------
    num : number
        The normalized number
    prefix : str
        One of the metric prefix strings. If `num` is already
        normalized, this is an empty string.
    factor : float
        A factor such that ``num * factor`` is the original input.
        If ``num`` is normalized, `factor` is 1.0.
    """
    power = floor(log10(abs(num)))
    if eng or power < -3 or power > 3:
        power = (power // 3) * 3
    prefix = _metric_prefixes[power][bool(long)]
    factor = 10**power
    return num / factor, prefix, factor



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 09 Jan 2021: Initial Coding

"""
Tools to help with threading.
"""


__all__ = ['Heartbeat']


from contextlib import contextmanager
from time import sleep
from threading import Thread


class Heartbeat(Thread):
    """
    A simple timer-like thread that emits a signal at fixed intervals.

    The thread can be stopped and paused using a context manager.
    """
    def __init__(self, emit, name='heartbeat', interval=1.0):
        """
        Construct a thread with the specified signal, name and interval.

        Parameters
        ----------
        emit : callable
            The no-arg callable to invoke to emit a signal.
        name : str, optional
            The name of the thread. The default is 'heartbeat'.
        interval : float, optional
            The interval from the end of one call to `emit` and the next
            invocation. Units are seconds. The default is 1.0.
        """
        super().__init__(name=name)
        self.running = False
        self.suspended = False
        self.emit = emit
        self.interval = float(interval)

    def start(self):
        """
        Start the thread.
        """
        self.running = True
        super().start()

    def stop(self):
        """
        Stop the heartbeat.

        The thread may not die until the current interval completes, but
        the signal will not be emitted again once this method is called.
        """
        self.running = False

    @contextmanager
    def suspend(self):
        """
        Context manager to temporarily suspend the heartbeat.

        Emission will stop when the manager enters, and resume on the
        next interval when it exits. Thread will die if an error occurs
        during suspension. Intervals are still timed while the thread is
        suspened.
        """
        self.suspended = True
        try:
            yield
        except:
            # Do not un-suspend on error: prevent last iteration from running
            self.running = False
            raise
        else:
            self.suspended = False

    def run(self):
        """
        Emit a signal immediately, and at somewhat regular intervals
        thereafter.
        """
        self.emit()
        while self.running:
            sleep(self.interval)
            if not self.suspended and self.running:
                self.emit()



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 27 Aug 2023: Initial Coding

"""
Tests for the :py:mod:`haggis.string_util` module.
"""


from pytest import raises

from ..string_util import horiz_cat, split_horiz_cat, parse_list, parse_pair


class TestParsePair:
    def test_trim(self):
        parser = parse_pair(trim=str.strip)
        assert parser(' 10 -12   ') == ('10', '12')

    def test_item_type(self):
        parser = parse_pair(item_type=int)
        assert parser('10-12') == (10, 12)
        with raises(ValueError,
                    match=r"invalid literal for int\(\) with base 10: 'x'"):
            parser('x-10')

    def test_sep_multi(self):
        parser = parse_pair(sep=[':', '-'])
        assert parser('a-x') == ('a', 'x')
        assert parser('10:(d)') == ('10', '(d)')
        with raises(ValueError, match='Pair requires 2 items, found 3'):
            parser('-8:12')
        with raises(ValueError, match='Pair requires 2 items, found 1'):
            parser('10_12')

    def test_inclusive_range(self):
        parser = parse_pair.inclusive_range()
        assert parser('10-12') == range(10, 13)
        with raises(ValueError, match='Pair requires 2 items, found 3'):
            parser('-8 - 12')

    def test_inclusive_slice(self):
        parser = parse_pair.inclusive_range(sep=':', output_type=slice)
        assert parser('10:12') == slice(10, 13)
        assert parser('-8 : 4') == slice(-8, 5)


class TestParseList:
    def test_number_list(self):
        parser = parse_list.number_list()
        assert parser('1,2,4-7,8') == [1, 2, (4, 7), 8]
        assert parser('1, 3, 5, 7') == [1, 3, 5, 7]
        assert parser('2') == [2]
        assert parser('10 - 12') == [(10, 12)]
        with raises(TypeError, match=r"Unable to parse '\[10:12\]' as any of "
                "{parse_pair\(sep='-', trim=None, item_type=int, "
                             "output_type=None\), int}"):
            parser('[10:12]')


class TestHoizCat:
    # Two equal columns
    sample_input = 'a\nabc\nab', '123\n\nx'

    # Three unequal columns
    missing_input = 'a\nabc\nxy\nz', 'b\nbc', 'c\ncde\nfg'

    def test_empty_input(self):
        assert horiz_cat(alignment='<', prefix='>', suffix='<') == ''
        with raises(IndexError, match=r'Number of headers \(1\) does not '
                                      r'match number of columns \(0\)'):
            horiz_cat(headers=['='])

    def test_empty_strings(self):
        for i in range(1, 4):
            s = [''] * i
            expected = f'>{"." * (i - 1)}<'
            assert horiz_cat(*s, sep='.',
                                   prefix='>', suffix='<') == expected

    def test_prefix(self):
        expected = '+a   123\n+abc    \n+ab  x  '
        assert horiz_cat(*self.sample_input, prefix='+') == expected

    def test_suffix(self):
        expected = 'a   123+\nabc    +\nab  x  +'
        assert horiz_cat(*self.sample_input, suffix='+') == expected

    def test_sep(self):
        expected = 'a  _123\nabc_   \nab _x  '
        assert horiz_cat(*self.sample_input, sep='_') == expected

    def test_sep_empty(self):
        expected2 = 'a  123\nabc   \nab x  '
        assert horiz_cat(*self.sample_input, sep='') == expected2

    def test_sep_not_str(self):
        with raises(AttributeError, match='\'NoneType\' object'):
            horiz_cat(*self.sample_input, sep=None)

    def test_linesep_default(self):
        expected = 'a   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input) == expected
        assert horiz_cat(*self.sample_input, linesep='\n') == expected

    def test_linesep(self):
        expected = 'a   123-x-abc    -x-ab  x  '
        assert horiz_cat(*self.sample_input, linesep='-x-') == expected

    def test_linesep_none(self):
        expected = ['a   123', 'abc    ', 'ab  x  ']
        assert horiz_cat(*self.sample_input, linesep=None) == expected

    def test_alignment_single(self):
        for alignments, expected in [
                (('<', 'left', 'LeFt'), 'a   123\nabc    \nab  x  '),
                (('>', 'right', 'rIgHt'), '  a 123\nabc    \n ab   x'),
                (('^', 'center', 'CENTER'), ' a  123\nabc    \nab   x '),
                ((None, 'none', 'None', ''), 'a 123\nabc \nab x')]:
            for alignment in alignments:
                assert horiz_cat(*self.sample_input,
                                       alignment=alignment) == expected

    def test_alignment_multiple(self):
        for alignments, expected in [
                (['^', '>'], ' a  123\nabc    \nab    x'),
                (list('><'), '  a 123\nabc    \n ab x  ')]:
           assert horiz_cat(*self.sample_input,
                                  alignment=alignments) == expected

    def test_alignment_invalid(self):
        with raises(ValueError):
            horiz_cat(*self.sample_input, alignment='no')
        with raises(TypeError, match='object of type \'bool\''):
            horiz_cat(*self.sample_input, alignment=True)
        with raises(IndexError, match=r'Number of alignments \(0\) does not '
                                      r'match number of columns \(2\)'):
            horiz_cat(*self.sample_input, alignment=[])
        with raises(IndexError, match=r'Number of alignments \(3\) does not '
                                      r'match number of columns \(2\)'):
            horiz_cat(*self.sample_input, alignment=list('<^>'))

    def test_missing_default(self):
        expected = 'a  .b .c  \nabc.bc.cde\nxy .  .fg \nz  .  .   '
        base_expected = 'a  .123\nabc.   \nab .x  '
        assert horiz_cat(*self.missing_input, sep='.') == expected
        assert horiz_cat(*self.sample_input, sep='.') == base_expected

    def test_missing(self):
        base_expected = 'a  .123\nabc.   \nab .x  '
        for missings, expected in [
                (('empty', 'EMPTY'),
                    'a  .b .c  \nabc.bc.cde\nxy .  .fg \nz  .  .   '),
                (('down', 'DoWn'),
                    'a  .  .   \nabc.  .c  \nxy .b .cde\nz  .bc.fg '),
                (('trunc', 'tRuNc'), 'a  .b .c  \nabc.bc.cde'),
                (('last', 'Last', -1, '-1'),
                    'a  .b .c  \nabc.bc.cde\nxy .bc.fg \nz  .bc.fg '),
                (('first', 'firST', 0, '0'),
                    'a  .b .c  \nabc.bc.cde\nxy .b .fg \nz  .b .c  '),
                ((1, '1'), 'a  .b .c  \nabc.bc.cde\nxy .bc.fg \nz  .bc.cde'),
                ((-2, '-2'),
                    'a  .b .c  \nabc.bc.cde\nxy .b .fg \nz  .b .cde')]:
            for missing in missings:
                assert horiz_cat(*self.missing_input, sep='.',
                                       missing=missing) == expected
                assert horiz_cat(*self.sample_input, sep='.',
                                       missing=missing) == base_expected

    def test_missing_invalid(self):
        with raises(ValueError, match='Invalid value for `missing`: \'sdkj\''):
            horiz_cat(*self.missing_input, sep='.', missing='sdkj')
        with raises(ValueError, match=r'Invalid value for `missing`: \[\]'):
            horiz_cat(*self.missing_input, sep='.', missing=[])
        with raises(ValueError, match='Invalid value for `missing`: None'):
            horiz_cat(*self.missing_input, sep='.', missing=None)
        for m in (2, -3, '2', '-3'):
            with raises(IndexError, match='list index out of range'):
                horiz_cat(*self.missing_input, sep='.', missing=m)

    def test_len_key(self):
        expected = 'a 123\nabc \nab x'
        assert horiz_cat(*self.sample_input,
                               len_key=lambda s: 3) == expected

    def test_titles_wider_str(self):
        expected = 'col1 col2\na    123 \nabc      \nab   x   '
        assert horiz_cat(*self.sample_input,
                               titles=['col1', 'col2']) == expected

    def test_titles_with_alignment(self):
        expected = ' X    Y\n a  123\nabc    \nab    x'
        assert horiz_cat(*self.sample_input, titles=list('XY'),
                               alignment=['^', '>']) == expected

    def test_titles_scalar(self):
        expected = 'ABC ABC\na   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input, titles='ABC') == expected

    def test_titles_empty(self):
        expected = '       \na   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input, titles=[''] * 2) == expected

    def test_titles_multiline(self):
        expected = 'A    B  \nCDEF Z  \na    123\nabc     \nab   x  '
        assert horiz_cat(*self.sample_input,
                               titles=['A\nCDEF', 'B\nZ']) == expected

    def test_titles_diff_heights(self):
        expected = '        \nCDEF Z  \na    123\nabc     \nab   x  '
        assert horiz_cat(*self.sample_input,
                               titles=['CDEF', '\nZ']) == expected

    def test_titles_invalid(self):
        with raises(IndexError, match=r'Number of titles \(0\) does not '
                                      r'match number of columns \(2\)'):
            horiz_cat(*self.sample_input, titles=[])
        with raises(IndexError, match=r'Number of titles \(3\) does not '
                                      r'match number of columns \(2\)'):
            horiz_cat(*self.sample_input, titles=list('abc'))

    def test_headers_one_char(self):
        expected = '=   =  \na   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input, headers='=') == expected

    def test_headers_longer_str(self):
        expected = '==== ====\na    123 \nabc      \nab   x   '
        assert horiz_cat(*self.sample_input, headers='====') == expected

    def test_headers_one_func(self):
        expected = '=== ===\na   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input,
                               headers='='.__mul__) == expected

    def test_headers_mixed_types(self):
        expected = '==== ===\na    123\nabc     \nab   x  '
        assert horiz_cat(*self.sample_input,
                               headers=['====', '='.__mul__]) == expected

    def test_headers_longer_func(self):
        expected = '===== =====\na   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input,
                               headers=lambda n: '=====') == expected

    def test_headers_shorter_func(self):
        expected = '== ==\na   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input,
                               headers=lambda n: '==') == expected

    def test_headers_empty_strs(self):
        expected = '       \na   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input,
                               headers=[''] * 2) == expected

    def test_headers_multiline(self):
        expected = '-\n- =  \na   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input,
                               headers=['-\n-', '=']) == expected

    def test_headers_invalid(self):
        with raises(IndexError, match=r'Number of headers \(0\) does not '
                                      r'match number of columns \(2\)'):
            horiz_cat(*self.sample_input, headers=[])
        with raises(IndexError, match=r'Number of headers \(3\) does not '
                                      r'match number of columns \(2\)'):
            horiz_cat(*self.sample_input, headers=list('abc'))

    def test_hsep_default(self):
        expected = '-  +-  \na  +123\nabc+   \nab +x  '
        assert horiz_cat(*self.sample_input, headers='-',
                               sep='+') == expected

    def test_hsep(self):
        expected = 'A   B  \n-  +-  \na   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input, headers='-',
                               titles=list('AB'), hsep='+') == expected

    def test_hsep_no_headers(self):
        expected = 'a   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input, sep=' ',
                               hsep='xxx') == expected

    def test_hsep_one_column_headers(self):
        expected = 'a  \nabc\nab '
        assert horiz_cat(self.sample_input[0], sep=' ',
                               hsep='xxx') == expected

    def test_hsep_with_len_key(self):
        expected = '-xxx-\na 123\nabc \nab x'
        assert horiz_cat(*self.sample_input, headers='-', sep=' ',
                               hsep='xxx', len_key=lambda s: 3) == expected

    def test_hsep_invalid(self):
        with raises(ValueError,
                    match='`hsep` and `sep` must be the same length'):
            horiz_cat(*self.sample_input, headers='-', sep=' ', hsep='')
        with raises(ValueError,
                    match='`hsep` and `sep` must be the same length'):
            horiz_cat(*self.sample_input, headers='-', sep=' ',
                            hsep='xxx')
        with raises(TypeError, match='object of type \'bool\''):
            horiz_cat(*self.sample_input, headers='-', sep=' ',
                            hsep=True)
        with raises(ValueError,
                    match='`hsep` and `sep` must be the same length'):
            horiz_cat(*self.sample_input, headers='-', sep=' ',
                            hsep=[1, 'a'])
        with raises(AttributeError,
                    match='\'list\' object has no attribute \'join\''):
            horiz_cat(*self.sample_input, headers='-', sep=' ', hsep=[1])

    def test_width_greater_scalar(self):
        expected = 'a    123 \nabc      \nab   x   '
        assert horiz_cat(*self.sample_input, width=4) == expected

    def test_width_smaller_scalar(self):
        expected = 'a   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input, width=2) == expected

    def test_width_mixed(self):
        expected = 'a   123 \nabc     \nab  x   '
        assert horiz_cat(*self.sample_input, width=[2, 4]) == expected

    def test_width_default(self):
        expected = 'a   123\nabc    \nab  x  '
        assert horiz_cat(*self.sample_input) == expected
        assert horiz_cat(*self.sample_input, width=None) == expected

    def test_width_none_mixed(self):
        expected1 = 'a   123 \nabc     \nab  x   '
        assert horiz_cat(*self.sample_input,
                               width=[None, 4]) == expected1
        expected2 = 'a    123\nabc     \nab   x  '
        assert horiz_cat(*self.sample_input,
                               width=[4, None]) == expected2

    def test_width_invalid(self):
        with raises(IndexError, match=r'Number of widths \(0\) does not '
                                      r'match number of columns \(2\)'):
            horiz_cat(*self.sample_input, width=[])
        with raises(IndexError, match=r'Number of widths \(3\) does not '
                                      r'match number of columns \(2\)'):
            horiz_cat(*self.sample_input, width=range(3))
        with raises(TypeError,
                    match='between instances of \'int\' and \'str\''):
            horiz_cat(*self.sample_input, width='xx')
        with raises(TypeError,
                    match='between instances of \'int\' and \'type\''):
            horiz_cat(*self.sample_input, width=[bool, 3])
        with raises(IndexError, match=r'Number of widths \(4\) does not '
                                      r'match number of columns \(2\)'):
            horiz_cat(*self.sample_input, width='None')

    def test_table(self):
        expected = '| Col1 |Col2 |\n|======+=====|\n|  a   | 123 |\n' \
                   '| abc  |     |\n|  ab  |  x  |'
        assert horiz_cat(*self.sample_input, prefix='|', suffix='|',
                               alignment='^', titles=['Col1', 'Col2'],
                               headers=['======', '='.__mul__], hsep='+',
                               sep='|', width=[6, 5]) == expected

class TestSplitHoizCat:
    # Two equal columns
    sample_input = ['a', 'abc', 'ab'], ['123', '', 'x']

    # Three unequal columns
    missing_input = ['a', 'abc', 'xy', 'z'], ['b', 'bc'], ['c', 'cde', 'fg']

    def test_empty_input(self):
        assert split_horiz_cat(alignment='<', prefix='>', suffix='<') == ''
        with raises(IndexError, match=r'Number of headers \(1\) does not '
                                      r'match number of columns \(0\)'):
            split_horiz_cat(headers=['='])

    def test_empty_strings(self):
        for i in range(1, 4):
            s = [['']] * i
            expected = f'>{"." * (i - 1)}<'
            assert split_horiz_cat(*s, sep='.',
                                   prefix='>', suffix='<') == expected

    def test_prefix(self):
        expected = '+a   123\n+abc    \n+ab  x  '
        assert split_horiz_cat(*self.sample_input, prefix='+') == expected

    def test_suffix(self):
        expected = 'a   123+\nabc    +\nab  x  +'
        assert split_horiz_cat(*self.sample_input, suffix='+') == expected

    def test_sep(self):
        expected = 'a  _123\nabc_   \nab _x  '
        assert split_horiz_cat(*self.sample_input, sep='_') == expected

    def test_sep_empty(self):
        expected2 = 'a  123\nabc   \nab x  '
        assert split_horiz_cat(*self.sample_input, sep='') == expected2

    def test_sep_not_str(self):
        with raises(AttributeError, match='\'NoneType\' object'):
            split_horiz_cat(*self.sample_input, sep=None)

    def test_linesep_default(self):
        expected = 'a   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input) == expected
        assert split_horiz_cat(*self.sample_input, linesep='\n') == expected

    def test_linesep(self):
        expected = 'a   123-x-abc    -x-ab  x  '
        assert split_horiz_cat(*self.sample_input, linesep='-x-') == expected

    def test_linesep_none(self):
        expected = ['a   123', 'abc    ', 'ab  x  ']
        assert split_horiz_cat(*self.sample_input, linesep=None) == expected

    def test_alignment_single(self):
        for alignments, expected in [
                (('<', 'left', 'LeFt'), 'a   123\nabc    \nab  x  '),
                (('>', 'right', 'rIgHt'), '  a 123\nabc    \n ab   x'),
                (('^', 'center', 'CENTER'), ' a  123\nabc    \nab   x '),
                ((None, 'none', 'None', ''), 'a 123\nabc \nab x')]:
            for alignment in alignments:
                assert split_horiz_cat(*self.sample_input,
                                       alignment=alignment) == expected

    def test_alignment_multiple(self):
        for alignments, expected in [
                (['^', '>'], ' a  123\nabc    \nab    x'),
                (list('><'), '  a 123\nabc    \n ab x  ')]:
           assert split_horiz_cat(*self.sample_input,
                                  alignment=alignments) == expected

    def test_alignment_invalid(self):
        with raises(ValueError):
            split_horiz_cat(*self.sample_input, alignment='no')
        with raises(TypeError, match='object of type \'bool\''):
            split_horiz_cat(*self.sample_input, alignment=True)
        with raises(IndexError, match=r'Number of alignments \(0\) does not '
                                      r'match number of columns \(2\)'):
            split_horiz_cat(*self.sample_input, alignment=[])
        with raises(IndexError, match=r'Number of alignments \(3\) does not '
                                      r'match number of columns \(2\)'):
            split_horiz_cat(*self.sample_input, alignment=list('<^>'))

    def test_missing_default(self):
        expected = 'a  .b .c  \nabc.bc.cde\nxy .  .fg \nz  .  .   '
        base_expected = 'a  .123\nabc.   \nab .x  '
        assert split_horiz_cat(*self.missing_input, sep='.') == expected
        assert split_horiz_cat(*self.sample_input, sep='.') == base_expected

    def test_missing(self):
        base_expected = 'a  .123\nabc.   \nab .x  '
        for missings, expected in [
                (('empty', 'EMPTY'),
                    'a  .b .c  \nabc.bc.cde\nxy .  .fg \nz  .  .   '),
                (('down', 'DoWn'),
                    'a  .  .   \nabc.  .c  \nxy .b .cde\nz  .bc.fg '),
                (('trunc', 'tRuNc'), 'a  .b .c  \nabc.bc.cde'),
                (('last', 'Last', -1, '-1'),
                    'a  .b .c  \nabc.bc.cde\nxy .bc.fg \nz  .bc.fg '),
                (('first', 'firST', 0, '0'),
                    'a  .b .c  \nabc.bc.cde\nxy .b .fg \nz  .b .c  '),
                ((1, '1'), 'a  .b .c  \nabc.bc.cde\nxy .bc.fg \nz  .bc.cde'),
                ((-2, '-2'),
                    'a  .b .c  \nabc.bc.cde\nxy .b .fg \nz  .b .cde')]:
            for missing in missings:
                assert split_horiz_cat(*self.missing_input, sep='.',
                                       missing=missing) == expected
                assert split_horiz_cat(*self.sample_input, sep='.',
                                       missing=missing) == base_expected

    def test_missing_invalid(self):
        with raises(ValueError, match='Invalid value for `missing`: \'sdkj\''):
            split_horiz_cat(*self.missing_input, sep='.', missing='sdkj')
        with raises(ValueError, match=r'Invalid value for `missing`: \[\]'):
            split_horiz_cat(*self.missing_input, sep='.', missing=[])
        with raises(ValueError, match='Invalid value for `missing`: None'):
            split_horiz_cat(*self.missing_input, sep='.', missing=None)
        for m in (2, -3, '2', '-3'):
            with raises(IndexError, match='list index out of range'):
                split_horiz_cat(*self.missing_input, sep='.', missing=m)

    def test_len_key(self):
        expected = 'a 123\nabc \nab x'
        assert split_horiz_cat(*self.sample_input,
                               len_key=lambda s: 3) == expected

    def test_titles_wider_str(self):
        expected = 'col1 col2\na    123 \nabc      \nab   x   '
        assert split_horiz_cat(*self.sample_input,
                               titles=['col1', 'col2']) == expected

    def test_titles_with_alignment(self):
        expected = ' X    Y\n a  123\nabc    \nab    x'
        assert split_horiz_cat(*self.sample_input, titles=list('XY'),
                               alignment=['^', '>']) == expected

    def test_titles_scalar(self):
        expected = 'ABC ABC\na   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input, titles='ABC') == expected

    def test_titles_empty(self):
        expected = '       \na   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input, titles=[''] * 2) == expected

    def test_titles_multiline(self):
        expected = 'A    B  \nCDEF Z  \na    123\nabc     \nab   x  '
        assert split_horiz_cat(*self.sample_input,
                               titles=['A\nCDEF', 'B\nZ']) == expected

    def test_titles_diff_heights(self):
        expected = '        \nCDEF Z  \na    123\nabc     \nab   x  '
        assert split_horiz_cat(*self.sample_input,
                               titles=['CDEF', '\nZ']) == expected

    def test_titles_invalid(self):
        with raises(IndexError, match=r'Number of titles \(0\) does not '
                                      r'match number of columns \(2\)'):
            split_horiz_cat(*self.sample_input, titles=[])
        with raises(IndexError, match=r'Number of titles \(3\) does not '
                                      r'match number of columns \(2\)'):
            split_horiz_cat(*self.sample_input, titles=list('abc'))

    def test_headers_one_char(self):
        expected = '=   =  \na   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input, headers='=') == expected

    def test_headers_longer_str(self):
        expected = '==== ====\na    123 \nabc      \nab   x   '
        assert split_horiz_cat(*self.sample_input, headers='====') == expected

    def test_headers_one_func(self):
        expected = '=== ===\na   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input,
                               headers='='.__mul__) == expected

    def test_headers_mixed_types(self):
        expected = '==== ===\na    123\nabc     \nab   x  '
        assert split_horiz_cat(*self.sample_input,
                               headers=['====', '='.__mul__]) == expected

    def test_headers_longer_func(self):
        expected = '===== =====\na   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input,
                               headers=lambda n: '=====') == expected

    def test_headers_shorter_func(self):
        expected = '== ==\na   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input,
                               headers=lambda n: '==') == expected

    def test_headers_empty_strs(self):
        expected = '       \na   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input,
                               headers=[''] * 2) == expected

    def test_headers_multiline(self):
        expected = '-\n- =  \na   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input,
                               headers=['-\n-', '=']) == expected

    def test_headers_invalid(self):
        with raises(IndexError, match=r'Number of headers \(0\) does not '
                                      r'match number of columns \(2\)'):
            split_horiz_cat(*self.sample_input, headers=[])
        with raises(IndexError, match=r'Number of headers \(3\) does not '
                                      r'match number of columns \(2\)'):
            split_horiz_cat(*self.sample_input, headers=list('abc'))

    def test_hsep_default(self):
        expected = '-  +-  \na  +123\nabc+   \nab +x  '
        assert split_horiz_cat(*self.sample_input, headers='-',
                               sep='+') == expected

    def test_hsep(self):
        expected = 'A   B  \n-  +-  \na   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input, headers='-',
                               titles=list('AB'), hsep='+') == expected

    def test_hsep_no_headers(self):
        expected = 'a   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input, sep=' ',
                               hsep='xxx') == expected

    def test_hsep_one_column_headers(self):
        expected = 'a  \nabc\nab '
        assert split_horiz_cat(self.sample_input[0], sep=' ',
                               hsep='xxx') == expected

    def test_hsep_with_len_key(self):
        expected = '-xxx-\na 123\nabc \nab x'
        assert split_horiz_cat(*self.sample_input, headers='-', sep=' ',
                               hsep='xxx', len_key=lambda s: 3) == expected

    def test_hsep_invalid(self):
        with raises(ValueError,
                    match='`hsep` and `sep` must be the same length'):
            split_horiz_cat(*self.sample_input, headers='-', sep=' ', hsep='')
        with raises(ValueError,
                    match='`hsep` and `sep` must be the same length'):
            split_horiz_cat(*self.sample_input, headers='-', sep=' ',
                            hsep='xxx')
        with raises(TypeError, match='object of type \'bool\''):
            split_horiz_cat(*self.sample_input, headers='-', sep=' ',
                            hsep=True)
        with raises(ValueError,
                    match='`hsep` and `sep` must be the same length'):
            split_horiz_cat(*self.sample_input, headers='-', sep=' ',
                            hsep=[1, 'a'])
        with raises(AttributeError,
                    match='\'list\' object has no attribute \'join\''):
            split_horiz_cat(*self.sample_input, headers='-', sep=' ', hsep=[1])

    def test_width_greater_scalar(self):
        expected = 'a    123 \nabc      \nab   x   '
        assert split_horiz_cat(*self.sample_input, width=4) == expected

    def test_width_smaller_scalar(self):
        expected = 'a   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input, width=2) == expected

    def test_width_mixed(self):
        expected = 'a   123 \nabc     \nab  x   '
        assert split_horiz_cat(*self.sample_input, width=[2, 4]) == expected

    def test_width_default(self):
        expected = 'a   123\nabc    \nab  x  '
        assert split_horiz_cat(*self.sample_input) == expected
        assert split_horiz_cat(*self.sample_input, width=None) == expected

    def test_width_none_mixed(self):
        expected1 = 'a   123 \nabc     \nab  x   '
        assert split_horiz_cat(*self.sample_input,
                               width=[None, 4]) == expected1
        expected2 = 'a    123\nabc     \nab   x  '
        assert split_horiz_cat(*self.sample_input,
                               width=[4, None]) == expected2

    def test_width_invalid(self):
        with raises(IndexError, match=r'Number of widths \(0\) does not '
                                      r'match number of columns \(2\)'):
            split_horiz_cat(*self.sample_input, width=[])
        with raises(IndexError, match=r'Number of widths \(3\) does not '
                                      r'match number of columns \(2\)'):
            split_horiz_cat(*self.sample_input, width=range(3))
        with raises(TypeError,
                    match='between instances of \'int\' and \'str\''):
            split_horiz_cat(*self.sample_input, width='xx')
        with raises(TypeError,
                    match='between instances of \'int\' and \'type\''):
            split_horiz_cat(*self.sample_input, width=[bool, 3])
        with raises(IndexError, match=r'Number of widths \(4\) does not '
                                      r'match number of columns \(2\)'):
            split_horiz_cat(*self.sample_input, width='None')

    def test_table(self):
        expected = '| Col1 |Col2 |\n|======+=====|\n|  a   | 123 |\n' \
                   '| abc  |     |\n|  ab  |  x  |'
        assert split_horiz_cat(*self.sample_input, prefix='|', suffix='|',
                               alignment='^', titles=['Col1', 'Col2'],
                               headers=['======', '='.__mul__], hsep='+',
                               sep='|', width=[6, 5]) == expected



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Recipes for common tasks that build on the type of thing normally found
in the builtin :py:mod:`os` module.
"""

__all__ = [
    'command_line', 'root_path', 'filter_copy', 'add_system_path',
    'chdir_context',
    'Tee', 'StdoutTee', 'StderrTee',
]


from contextlib import contextmanager
from functools import partial
from os import chdir, environ, getcwd, sep, pathsep
from os.path import abspath, splitdrive
import sys

from .files import open_file
from .string_util import hasspace


def command_line(exec=None, args=None, quote='"'):
    """
    Reconstruct a command line based on the specified executable
    `exec` and iterable of arguments `args`.

    `exec` defaults to :py:obj:`sys.executable` and `args` defaults
    to :py:obj:`sys.argv`.
    """
    def argue(arg):
        """
        Coerce the input to a string and puts quotes around it if it
        contains whitespace.
        """
        arg = str(arg)
        if hasspace(arg):
            return '{}{}{}'.format(quote, arg, quote)
        return arg

    if exec is None:
        exec = sys.executable
    if args is None:
        args = sys.argv

    return '{} {}'.format(argue(exec), ' '.join(map(argue, args)))


def root_path(file=None):
    """
    Retrieve the root file system for the given file, or for the whole
    OS.

    On Windows this will be a drive letter followed by a backslash. On
    most Unix variants, this will just be a slash.
    """
    if file:
        # Modified from http://stackoverflow.com/a/12041606/2988730
        drive, _ = splitdrive(file)
        return drive + sep
    # Courtesy of http://stackoverflow.com/a/22255432/2988730
    return abspath(sep)


def filter_copy(src, dest, hook=None, *, encoding=None, strip_newlines=False):
    """
    Copy a file line by line with optional processing of the lines.

    `src` and `dest` can be either file-like objects or strings or file
    descriptors. If `src` is file-like, it will only be copied from the
    current position of the cursor. If `dest` is file-like, it will be
    appended to or overwritten from the current position of the cursor.
    Otherwise, it will be truncated.

    `hook` is a function that accepts a line from `src` and returns the
    modified line to write into `dest`. The default value of
    :py:obj:`None` is equivalent to a pass-thru like ``lambda x: x``. A
    return value of :py:obj:`None` means to print nothing to the file.
    This is not the same as an empty string if ``strip_newlines=True``.

    If `strip_newlines` is :py:obj:`False` (the default), the input to
    `hook` will contain the trailing newline characters. Whether or not
    the output does is entirely up to the implementation, but one will
    not be automatically appended. If `strip_newlines` is
    :py:obj:`True`, the input to `hook` will not contain the trailing
    newline and one will be appended to the output if the input
    contained one.
    """
    with open_file(src, 'rt', encoding=encoding) as src_file, \
         open_file(dest, 'wt', encoding=encoding) as dest_file:
        while True:
            line = src_file.readline()
            if not line:
                break
            if strip_newlines and line[-1] == '\n':
                eol = line[-1]
                line = line[:-1]
            else:
                eol = ''
            if hook:
                line = hook(line)
            if line is not None:
                dest_file.write(line)
                dest_file.write(eol)


def add_system_path(*paths, append=True, var='PATH'):
    """
    Extends the :envvar:`PATH` environment variable with the specified
    sequence of additional elements.

    Elements are only added if they are not already present in the
    existing path. This function only does literal comparison and
    append. It does not account for environment variable expansion or
    anything like that.
    """
    env_path = environ.get(var, '').split(pathsep)

    add = env_path.append if append else partial(env_path.insert, 0)

    for path in paths:
        if path not in env_path:
            add(str(path))

    environ[var] = pathsep.join(filter(None, env_path))


class Tee:
    """
    An output stream that directs output to two different streams.

    This class provides a :py:meth:`write` and :py:meth:`flush` methods.
    Since it is intended to be used with :py:obj:`sys.stdout` and
    :py:obj:`sys.stderr`, it also provides an :py:meth:`isatty` method,
    which always returns :py:obj:`False`.

    .. py:attribute:: s1

       The first stream to write to in the tee.

    .. py:attribute:: s2

       The second stream to write to in the tee.

    No checking is done on the streams, e.g., to make sure that they are
    opened with the same mode, etc.

    .. todo:: Add the proper mixins/ABCs from io package.
    """
    def __init__(self, stream1, stream2):
        """
        Initializes a tee to the specified streams.
        """
        self.s1 = stream1
        self.s2 = stream2

    def __repr__(self):
        return '{type}(s1={self.s1!r}, s2={self.s2!r})'.format(
            type=type(self).__qualname__, self=self
        )

    def write(self, string):
        """
        Write the output to both teed streams.

        This method does not return anything.
        """
        self.s1.write(string)
        self.s2.write(string)

    def flush(self):
        """
        Attempt to flush both teed streams.

        Streams are only flushed if they have a callable ``flush``
        method. Closed streams will not be flushed.
        """
        self._call_method(self.s1, 'flush')
        self._call_method(self.s2, 'flush')

    def isatty(self):
        """
        Always return :py:obj:`False` to indicate redirection.
        """
        return False

    def close(self):
        """
        Closes either of the underlying streams that is not a TTY.

        Streams with no callable ``isatty`` attribute are closed if they
        have a callable ``close`` attribute.
        """
        def close_not_tty(obj):
            if not self._call_method(obj, 'isatty'):
                self._call_method(obj, 'close')

        close_not_tty(self.s1)
        close_not_tty(self.s2)

    @staticmethod
    def _call_method(obj, name):
        """
        Checks if a (file) object has a named method and returns the
        result of calling it.

        If the attribute does not exist or is not :py:func:`callable`,
        return :py:obj:`None`. Methods are assumed to be no-arg.

        Methods are expected to raise a :py:exc:`ValueError` if the file
        is closed. The error will be ignored and the return value will
        be :py:obj:`None`.
        """
        meth = getattr(obj, name, None)
        if meth is None or not callable(meth):
            return None
        try:
            return meth()
        except ValueError:
            # Raised to indicate that file is already closed:
            # https://stackoverflow.com/a/11859306/2988730
            return None


class StdoutTee(Tee):
    """
    Tees output to :py:obj:`sys.stdout` and another stream.

    This class replaces :py:obj:`sys.stdout` if used as a context
    manager. It retains a reference to the original stream, which it
    replaces on exit.
    """
    def __init__(self, stream):
        """
        Initialize a tee with `stream` and :py:obj:`sys.stdout`.
        """
        super().__init__(stream, sys.stdout)

    def __enter__(self):
        """
        Replace :py:obj:`sys.stdout` with this object.

        Return this object.
        """
        sys.stdout = self
        return self

    def __exit__(self, *args, **kwargs):
        """
        Replace the current :py:obj:`sys.stdout` with the original and
        close the other stream.
        """
        sys.stdout = self.s2
        self.close()


class StderrTee(Tee):
    """
    Tees output to :py:obj:`sys.stderr` and another stream.

    This class replaces :py:obj:`sys.stderr` if used as a context
    manager. It retains a reference to the original stream, which it
    replaces on exit.
    """
    def __init__(self, stream):
        """
        Initialize a tee with `stream` and :py:obj:`sys.stderr`.
        """
        super().__init__(stream, sys.stderr)

    def __enter__(self):
        """
        Replace :py:obj:`sys.stderr` with this object.

        Return this object.
        """
        sys.stderr = self
        return self

    def __exit__(self, *args, **kwargs):
        """
        Replace the current :py:obj:`sys.stderr` with the original and
        close the other stream.
        """
        sys.stderr = self.s2
        self.close()


@contextmanager
def chdir_context(path, current=None):
    """
    A context manager that changes the current directory using
    :py:func:`os.chdir`.

    The `current` directory is reinstated once the manager exits.

    Parameters
    ----------
    path : path-like or file-descriptor
        The directory to temporarily change to. Any argument that is
        valid for :py:func:`os.chdir` is valid here.
    current : path-like or file-descriptor or None
        The directory to return to once the context manager exits. If
        omitted or :py:obj:`None`, the current working directory as
        returned by :py:func:`os.getcwd` is used. As with `path`, the
        argument must be valid for :py:func:`os.chdir`.
    """
    if current is None:
        current = getcwd()

    try:
        chdir(path)
        yield
    finally:
        chdir(current)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
Setup script for building and installing haggis.
"""

from os.path import dirname, join
import sys

from setuptools import setup


DIST_NAME = 'haggis'

LICENSE = 'GNU Affero General Public License v3 or later (AGPLv3+)'
DESCRIPTION = 'General purpose utility library'

AUTHOR = 'Joseph R. Fox-Rabinovitz'
AUTHOR_EMAIL = 'jfoxrabinovitz@gmail.com'

MAINTAINER = 'Joseph R. Fox-Rabinovitz'
MAINTAINER_EMAIL = 'jfoxrabinovitz@gmail.com'

CLASSIFIERS = [
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Other Audience',
    'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX :: Linux',
    'Operating System :: Unix',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Topic :: Software Development :: Libraries',
    'Topic :: System :: Console Fonts',
    'Topic :: System :: Logging',
    'Topic :: Terminals',
    'Topic :: Text Processing',
    'Topic :: Text Processing :: General',
    'Topic :: Text Processing :: Markup',
    'Topic :: Text Processing :: Markup :: LaTeX',
    'Topic :: Text Processing :: Markup :: XML',
    'Topic :: Utilities',
]


COMMANDS = {}


try:
    from sphinx.setup_command import BuildDoc
    COMMANDS['build_sphinx'] = BuildDoc
except ImportError:
    pass


def import_file(name, location):
    """
    Imports the specified python file as a module, without explicitly
    registering it to `sys.modules`.

    While haggis uses Python 3 features, you are free to try to install
    it in a Python 2 environment.
    """
    if sys.version_info[0] == 2:
        # Python 2.7-
        from imp import load_source
        mod = load_source(name, location)
    elif sys.version_info < (3, 5, 0):
        # Python 3.4-
        from importlib.machinery import SourceFileLoader
        mod = SourceFileLoader(name, location).load_module()
    else:
        # Python 3.5+
        from importlib.util import spec_from_file_location, module_from_spec
        spec = spec_from_file_location(name, location)
        mod = module_from_spec(spec)
        spec.loader.exec_module(mod)
    return mod


def version_info():
    """
    Jump through some hoops to import version.py for the different
    versions of Python.

    https://stackoverflow.com/a/67692/2988730
    """
    location = join(dirname(__file__) or '.', 'src', 'haggis', 'version.py')
    mod = import_file('version', location)
    return mod.__version__


def long_description():
    """
    Reads in the README and CHANGELOG files, separated by two
    newlines.
    """
    with open('README.md') as readme, open('CHANGELOG') as changes:
        return '%s\n\n%s' % (readme.read(), changes.read())


if __name__ == '__main__':
    setup(
        name=DIST_NAME,
        version=version_info(),
        license=LICENSE,
        description=DESCRIPTION,
        long_description_content_type='text/markdown',
        long_description=long_description(),
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        classifiers=CLASSIFIERS,
        url='https://gitlab.com/madphysicist/haggis',
        project_urls={
            'Bugs': 'https://gitlab.com/madphysicist/haggis/-/issues',
            'Documentation': 'https://haggis.readthedocs.io/en/stable/',
        },
        packages=[
            'haggis',
            'haggis.files',
            'haggis.tests',
            'haggis.files.tests',
        ],
        package_dir={'': 'src'},
        package_data={'haggis.files': ['_resources/*.XSL'],
                      'haggis.tests': ['data/*']},
        install_requires=['numpy >= 1.10'],
        extras_require={
            'all': [
                'astropy >= 3.0',
                'colorama >= 0.3;platform_system=="Windows"',
                'matplotlib >= 1.5',
                'openpyxl >= 2.4.8',
                'python-docx >= 0.8.5',
                'scipy > 0.16',
            ],
            'docx': ['python-docx >= 0.8.5'],
            'latex': [],
            'pdf': [],
            'ps': [],
            'plot': ['matplotlib >= 1.5'],
            'scio': ['scipy > 0.16', 'astropy >= 3.0'],
            'term': ['colorama >= 0.3;platform_system=="Windows"'],
            'xlsx': ['openpyxl >= 2.4.8'],
        },
        provides=['haggis'],
        tests_require=['pytest', 'pytest-cov'],
        data_files = [('', ['LICENSE', 'README.md', 'CHANGELOG'])],
        cmdclass=COMMANDS,
    )



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 19 May 2021: Initial Coding

"""
Utilities for interacting with the file system with an optional
graphical user interface.

The GUI package is chosen based on a list of supported selections,
defaulting to tkinter. Currently, only PyQt and tkinter are supported.
"""

__all__ = ['get_existing_file']


from os.path import isdir, isfile
from warnings import warn


_UI_ORDER = ['pyqt', 'tkinter']


def _iter(i):
    if isinstance(i, str):
        return i,
    return i


for name in _UI_ORDER:
    try:
        if name == 'pyqt':
            try:
                from PyQt5.QtWidgets import QFileDialog
            except ImportError:
                from PyQt4.QtGui import QFileDialog

            def _fix(ext):
                """
                Convert extensions with missing '*.' into canonical form.
                """
                ext = str(ext)
                if ext.startswith('*.') or ext == '*':
                    return ext
                elif ext.startswith('.'):
                    return '*' + ext
                elif ext.startswith('*'):
                    return ext[0] + '.' + ext[1:]
                return '*.' + ext

            def _filters(seq, sep):
                """
                Join all fixed extensions in `seq` with `sep`.
                """
                return sep.join(map(_fix, _iter(seq)))

            def _filt2str(item):
                """
                Create a Qt-compatible filter string based on a 2-element
                sequence containing a label and sequence of extensions.
                """
                return f'{item[0]} ({_filters(item[1], " ")})'

            def _filt2qt(filt, sel):
                """
                Convert a filter from `open_file` to Qt-suitable format.
                """
                if not filt:
                    return None, None

                try:
                    filtee = filt.items()
                    s = _filt2str((sel, filt[sel]))
                except AttributeError:
                    filtee = filt
                    s = _filt2str(filt[sel])
                return ';;'.join(map(_filt2str, filtee)), s

            def open_file(dir, title, filters=(), sel=0):
                """
                Qt implementation of the `open_file` interface.
                """
                filters, sel = _filt2qt(filters, sel)
                name, _ = QFileDialog.getOpenFileName(
                    None, title, dir, filters, sel
                )
                return name

        elif name == 'tkinter':
            from tkinter import Tk
            from tkinter.filedialog import askopenfilename

            def _fix(ext):
                """
                Convert extensions with missing '*.' into canonical form.
                """
                ext = str(ext)
                if ext.startswith('*.') or ext == '*':
                    return ext
                elif ext.startswith('.'):
                    return '*' + ext
                elif ext.startswith('*'):
                    return ext[0] + '.' + ext[1:]
                return '*.' + ext

            def _filt2tk(filt):
                """
                Convert a filter from `open_file` to Tk-suitable format.
                """
                if not filt:
                    return []

                try:
                    filtee = filt.items()
                except AttributeError:
                    filtee = filt
                return [(key, _fix(elem)) for key, value in filtee for elem in _iter(value)]

            def open_file(dir, title, filters=(), sel=0):
                """
                Tkinter implementation of the `open_file` interface.
                """
                filters = _filt2tk(filters)
                Tk().withdraw()
                return askopenfilename(
                    parent=None, title=title, initialdir=dir,
                    filetypes=filters
                )
    except ImportError:
        continue
    else:
        break
else:
    warn('No GUI package imported. All functions will fail.')

    def open_file(*args, **kwargs):
        raise TypeError('No supporting GUI package found')


def get_existing_file(filename=None, title='Open', filters=None, sel=0):
    """
    Return the name of an existing file.

    The file can be opened for reading unless permissions intervene.

    Parameters
    ----------
    filename : str, optional
        The initial file name to check. If the name exists and is a
        file, a GUI will not be displayed. If it is a directory, it will
        be used as the starting point in the GUI. The default is None.
    title : str or None
        An optional title for the dialog that will be displayed if
        `filename` does not exist.
    filters : sequence[str or tuple] or None
        A sequence or mapping of filename filters. Sequences must
        consist of two-tuples with a filter name and a list of
        extensions. Mappings contain a list of extensions as values. If
        None, no filtering will be done in the dialog. Filter extensions
        may contain a leading ``'*.'``, but are not required to contain
        either character.
    sel : str or int
        An optional key into `filters`, ignored if `filters` is None.

    Returns
    -------
    filename : str or None
        The name of an existing file, or None if the user changes their
        mind.
    """
    if filename is None:
        start = '.'
    else:
        if isfile(filename):
            return filename
        if isdir(filename):
            start = filename
        else:
            start = '.'

    return open_file(dir=start, title=title, filters=filters, sel=sel)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 27 Jan 2022: Initial Coding

"""
Tests for the :py:mod:`haggis.load` module.
"""


from os.path import abspath, dirname, join
import sys
from types import FunctionType, LambdaType, ModuleType

import pytest

from ..load import load_module, module_as_dict
from ..os import chdir_context


DATA_DIR = join(dirname(abspath(__file__)), 'data')


class Container:
    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)


class TestLoadModule:
    def setup_method(self):
        if 'module' in sys.modules:
            del sys.modules['module']

    def test_load(self):
        mod = load_module(join(DATA_DIR, 'module.txt'))
        assert mod.__name__ == 'module'
        assert hasattr(mod, 'a')
        assert hasattr(mod, 'b')
        assert hasattr(mod, 'c')
        assert hasattr(mod, '__d__')
        assert hasattr(mod, 'nested')
        assert 'module' not in sys.modules

    @pytest.mark.parametrize(('name', 'outname'),
            [(None, 'module'), ('', 'module'), ('testmodule', 'testmodule')])
    def test_name(self, name, outname):
        mod = load_module(join(DATA_DIR, 'module.txt'), name=name)
        assert mod.__name__ == outname
        assert hasattr(mod, 'a')
        assert hasattr(mod, 'b')
        assert hasattr(mod, 'c')
        assert hasattr(mod, '__d__')
        assert hasattr(mod, 'nested')
        assert outname not in sys.modules

    def test_sys(self):
        name = 'very_unlikely_test_module'
        assert name not in sys.modules
        try:
            mod = load_module(join(DATA_DIR, 'module.txt'), name=name, sys_module=True)
            assert mod.__name__ == name
            assert hasattr(mod, 'a')
            assert hasattr(mod, 'b')
            assert hasattr(mod, 'c')
            assert hasattr(mod, '__d__')
            assert hasattr(mod, 'nested')
            assert name in sys.modules
        finally:
            if name in sys.modules:
                del sys.modules[name]

    def test_injection(self):
        inject = Container(x=1)
        with pytest.raises(NameError):
            load_module(join(DATA_DIR, 'injection_module.txt'),
                        injection=inject, injection_var=None)

        mod = load_module(join(DATA_DIR, 'injection_module.txt'),
                          injection=inject, injection_var='injected')
        assert mod.injected is inject
        assert mod.a == 1


class TestModuleAsDict:
    def test_basic(self):
        config = module_as_dict(join(DATA_DIR, 'module.txt'))
        assert len(config) == 4
        assert config['a'] == 1
        assert isinstance(config['b'], FunctionType)
        assert isinstance(config['c'], type)
        assert '__d__' not in config
        assert 'e' not in config

    def test_include(self):
        with chdir_context(DATA_DIR):
            config = module_as_dict('module.txt', include_var='nested')
        assert len(config) == 5
        assert config['a'] == 1
        assert isinstance(config['b'], FunctionType)
        assert isinstance(config['c'], type)
        assert '__d__' not in config
        assert isinstance(config['d'], type)
        assert 'e' not in config
        

    def test_bfs(self):
        with chdir_context(DATA_DIR):
            config = module_as_dict('bfs_module_0.txt')
        assert len(config) == 4
        assert config['a'] == 1
        assert config['b'] == 2
        assert config['c'] == 3
        assert config['d'] == 42

    def test_skip_dunder(self):
        config = module_as_dict(join(DATA_DIR, 'module.txt'),
                                skip_dunder=False)
        assert config['a'] == 1
        assert isinstance(config['b'], FunctionType)
        assert isinstance(config['c'], type)
        assert isinstance(config['__d__'], LambdaType)
        assert 'e' not in config

    def test_skip_modules(self):
        config = module_as_dict(join(DATA_DIR, 'module.txt'),
                                skip_modules=False)
        assert len(config) == 5
        assert config['a'] == 1
        assert isinstance(config['b'], FunctionType)
        assert isinstance(config['c'], type)
        assert '__d__' not in config
        assert isinstance(config['e'], ModuleType)

    def test_skip_classes(self):
        config = module_as_dict(join(DATA_DIR, 'module.txt'),
                                skip_classes=True)
        assert len(config) == 3
        assert config['a'] == 1
        assert isinstance(config['b'], FunctionType)
        assert 'c' not in config
        assert '__d__' not in config
        assert 'e' not in config

    def test_skip_function(self):
        config = module_as_dict(join(DATA_DIR, 'module.txt'),
                                skip_functions=True, skip_dunder=False)
        assert config['a'] == 1
        assert 'b' not in config
        assert isinstance(config['c'], type)
        assert '__d__' not in config
        assert 'e' not in config

    def test_injection(self):
        with chdir_context(DATA_DIR):
            with pytest.raises(NameError):
                module_as_dict('injection_module.txt', include_var=None)
            with pytest.raises(AttributeError):
                module_as_dict('injection_module.txt',
                               injection_var='injected', include_var=None)
            config = module_as_dict('injection_module.txt',
                                    injection_var='injected',
                                    injection=Container(x=1),
                                    include_var=None)
            assert config['a'] == 1

    def test_recurse_injection(self):
        with chdir_context(DATA_DIR):
            with pytest.raises(NameError):
                module_as_dict('injection_module.txt',
                               injection=Container(x=1, y=2),
                               injection_var='injected',
                               recurse_injection=False)
            with pytest.raises(AttributeError):
                module_as_dict('injection_module.txt',
                               injection=Container(x=1),
                               injection_var='injected')
            config = module_as_dict('injection_module.txt',
                                    injection=Container(x=1, y=2),
                                    injection_var='injected')
            assert config['a'] == 1
            assert config['b'] == 2
